from pytezos.operation.forge import forge_operation

hard_gas_limit_per_operation = 1040000
hard_storage_limit_per_operation = 60000
minimal_fees = 100
minimal_nanotez_per_byte = 1
minimal_nanotez_per_gas_unit = .1


def calculate_fee(content: dict, consumed_gas: int, extra_size: int, reserve=10) -> int:
    """ Calculate minimal required operation fee.

    :param content: operation content {..., "kind": "transaction", ... }
    :param consumed_gas: amount of gas consumed during the simulation (dry-run)
    :param extra_size: size of the additional operation data (branch, etc)
    :param reserve: safe reserve, just in case
    """
    size = len(forge_operation(content)) + extra_size
    fee = minimal_fees \
        + minimal_nanotez_per_byte * size \
        + int(minimal_nanotez_per_gas_unit * consumed_gas)
    return fee + reserve


def default_fee(content) -> int:
    """ Take hard gas limit instead of precise amount (no simulation) and calculate fee.

    :param content: operation content {..., "kind": "transaction", ... }
    """
    return calculate_fee(
        content=content,
        consumed_gas=default_gas_limit(content),
        extra_size=32 + 64 + 3 * 3  # branch, signature, fee:gas_limit:storage_limit mutez values (+3 bytes)
    )


def default_gas_limit(content) -> int:
    """ Get default gas limit by operation kind.

    :param content: operation content {..., "kind": "transaction", ... }
    """
    values = {
        'reveal': 10000,
        'delegation': 10000,
        'origination': hard_gas_limit_per_operation if content.get('script') else 10000,
        'transaction': hard_gas_limit_per_operation if content.get('destination', '').startswith('KT') else 10207
    }
    return values.get(content['kind'])


def default_storage_limit(content):
    """ Get default storage limit by operation kind.

    :param content: operation content {..., "kind": "transaction", ... }
    """
    values = {
        'reveal': 0,
        'delegation': 0,
        'origination': hard_storage_limit_per_operation if content.get('script') else 10207,
        'transaction': hard_storage_limit_per_operation if content.get('destination', '').startswith('KT') else 257
    }
    return values.get(content['kind'])



from os.path import dirname

from pytezos.rpc.node import RpcNode
from pytezos.rpc.docs import rpc_docs
from pytezos.tools.docstring import get_attr_docstring, get_class_docstring, InlineDocstring


def format_docstring(class_type, query_path):
    res = ['']
    methods = {
        'GET': '()',
        'POST': '.post()',
        'PUT': '.put()',
        'DELETE': '.delete()'
    }
    rpc_doc = rpc_docs.get(query_path, {})

    for method, func in methods.items():
        if method in rpc_doc:
            docstring = get_attr_docstring(class_type, method.lower())
            if not docstring:
                docstring = f'\n{rpc_doc[method]["descr"]}\n'
                for arg in rpc_doc[method]['args']:
                    docstring += f':param {arg["name"]}: {arg["descr"]}\n'
                docstring += f':returns: {rpc_doc[method]["ret"]}\n'

            res.append(f'{func}{docstring}')

    if 'item' in rpc_doc:
        docstring = get_attr_docstring(class_type, '__getitem__')
        if not docstring:
            item = rpc_doc["item"]
            docstring = f'\n:param {item["name"]}: {item["descr"]}\n:returns: Child element\n'

        res.append(f'[]{docstring}')

    if 'props' in rpc_doc:
        properties = rpc_doc['props']
        docstring = '\n'.join(map(lambda x: f'.{x}', properties))
        res.append(f'RPC endpoints\n{docstring}\n')
    else:
        properties = list()

    helpers = get_class_docstring(
        class_type=class_type,
        attr_filter=lambda x: not x.startswith('_')
                              and x not in properties
                              and x.upper() not in methods
                              and x != 'path'
    )
    if helpers:
        res.append(f'Helpers\n{helpers}')

    return '\n'.join(res)


class RpcQuery(metaclass=InlineDocstring):
    __extensions__ = dict()

    @classmethod
    def __init_subclass__(cls, path='', **kwargs):
        super().__init_subclass__(**kwargs)
        if isinstance(path, list):
            for sub_path in path:
                cls.__extensions__[sub_path] = cls
        else:
            cls.__extensions__[path] = cls

    def __init__(self, node: RpcNode, path='', caching=False, params=None, timeout=None):
        self.node = node
        self._wild_path = path
        self._caching = caching
        self._timeout = timeout
        self._params = params or list()
        self.__doc__ = format_docstring(self.__class__, self._wild_path or '/')

    def __repr__(self):
        res = [
            super(RpcQuery, self).__repr__(),
            '\nProperties',
            f'.path  # {self.path or "/"}{" (cached)" if self._caching else ""}',
            f'.node  # {self.node.uri} ({self.node.network})',
            self.__doc__
        ]
        return '\n'.join(res)

    def _spawn_query(self, wild_path, params):
        child_class = self.__extensions__.get(wild_path, RpcQuery)
        return child_class(
            path=wild_path,
            node=self.node,
            caching=self._caching,
            params=params
        )

    @property
    def path(self):
        return self._wild_path.format(*self._params)

    def __call__(self, **params):
        return self.node.get(
            path=self.path,
            params=params,
            caching=self._caching
        )

    def _getitem(self, item):
        return self._spawn_query(
            wild_path=self._wild_path + '/{}',
            params=self._params + [item]
        )

    def __getattr__(self, attr):
        if not attr.startswith('_'):
            if attr in {'main', 'test', 'head', 'genesis'}:
                return self._getitem(attr)
            else:
                return self._spawn_query(
                    wild_path=f'{self._wild_path}/{attr}',
                    params=self._params
                )
        raise AttributeError(attr)

    def __getitem__(self, child_id):
        return self._getitem(child_id)

    def _get(self, params=None):
        return self.node.get(
            path=self.path,
            params=params,
            caching=self._caching,
            timeout=self._timeout
        )

    def _post(self, json=None, params=None):
        return self.node.post(
            path=self.path,
            params=params,
            json=json,
            caching=self._caching
        )

    def _put(self, params=None):
        return self.node.put(
            path=self.path,
            params=params
        )

    def _delete(self, params=None):
        return self.node.delete(
            path=self.path,
            params=params
        )

    @property
    def _parent(self):
        dir_path = dirname(self._wild_path)
        return self._spawn_query(
            wild_path=dir_path,
            params=self._params[:dir_path.count('{}')]
        )



import re
from copy import deepcopy
from pprint import pformat
from typing import List

from pytezos.crypto import blake2b
from pytezos.encoding import base58_encode
from pytezos.repl.types import StackItem, assert_stack_item
from pytezos.michelson.converter import micheline_to_michelson
from pytezos.repl.big_map import BigMapPool


class DummyGen:

    def __init__(self):
        self.index = 0
        self.self = self.get_fresh_address()

    def get_fresh_address(self):
        nonce = b'\x00' * 32 + self.index.to_bytes(4, 'big')
        nonce_hash = blake2b(data=nonce, digest_size=20).digest()
        res = base58_encode(nonce_hash, b'KT1').decode()
        self.index += 1
        return res


class Context:

    def __init__(self):
        self.stack = []
        self.meta = {}
        self.dummy_gen = DummyGen()
        self.big_maps = BigMapPool()
        self.debug = True
        self.stdout = list()
        self.protected = 0
        self.pushed = 0

    def __deepcopy__(self, memodict={}):
        ctx = Context()
        ctx.stack = deepcopy(self.stack)
        ctx.meta = deepcopy(self.meta)
        ctx.dummy_gen = deepcopy(self.dummy_gen)
        ctx.big_maps = deepcopy(self.big_maps)
        ctx.debug = self.debug
        ctx.protected = self.protected
        ctx.pushed = self.pushed
        return ctx

    def spawn(self, stack):
        ctx = Context()
        ctx.stack = stack
        ctx.meta = self.meta
        ctx.dummy_gen = self.dummy_gen
        ctx.big_maps = self.big_maps
        ctx.debug = self.debug
        ctx.stdout = self.stdout
        return ctx

    def reset(self):
        self.stdout = []
        self.pushed = False

    def protect(self, count: int):
        assert len(self.stack) >= count, f'got {len(self.stack)} items, wanted to protect {count}'
        self.protected += count
        self.print(f'protect {count} item(s)')

    def restore(self, count: int):
        assert self.protected >= count, f'wanted to restore {count}, only {self.protected} protected'
        self.print(f'restore {count} item(s)')
        self.protected -= count

    def push(self, item: StackItem, annots=None, move=False):
        assert_stack_item(item)
        self.stack.insert(self.protected, item if move else item.rename(annots))
        self.pushed = True
        self.print(f'push {repr(item)}')

    def peek(self):
        assert len(self.stack) > 0, 'stack is empty'
        return self.stack[self.protected]

    def pop(self, count: int) -> List[StackItem]:
        assert len(self.stack) - self.protected >= count, \
            f'got {len(self.stack) - self.protected} items, requested {count} '
        res = [self.stack.pop(self.protected) for _ in range(count)]
        if count <= 3:
            body = ', '.join(map(repr, res))  # TODO: restrict line length
        else:
            body = f'{count} items'
        self.print(f'pop {body}')
        return res

    def pop1(self):
        res = self.pop(count=1)
        return res[0]

    def pop2(self):
        return tuple(self.pop(count=2))

    def pop3(self):
        return tuple(self.pop(count=3))

    def get(self, key, default=None):
        return self.meta.get(key, default)

    def set(self, key, value):
        self.meta[key] = value
        if key in ['parameter', 'storage', 'code', 'STORAGE']:
            self.print(micheline_to_michelson({"prim": key, "args": [value]}, inline=True))
        else:
            self.print(f'set {key}={repr(value)}')

    def unset(self, key):
        if key in self.meta:
            del self.meta[key]
            self.print(f'unset {key}')

    def drop_all(self):
        self.stack.clear()
        self.protected = 0
        self.print(f'drop all')

    def dump(self, count: int):
        if len(self.stack) > 0:
            count = min(count, len(self.stack))
            return self.stack[:count]

    def print(self, message):
        if self.debug:
            self.stdout.append({'action': 'event', 'text': message})

    def printf(self, template: str):
        def format_stack_item(match):
            i = int(match.groups()[0])
            assert i < len(self.stack), f'requested {i}th element, got only {len(self.stack)} items'
            return repr(self.stack[i])

        message = re.sub(r'\{(\d+)\}', format_stack_item, template)
        self.stdout.append({'action': 'message', 'text': message})

    def begin(self, prim=None):
        self.pushed = False
        if self.debug:
            self.stdout.append({'action': 'begin', 'prim': prim})

    def end(self):
        if self.debug:
            self.stdout.append({'action': 'end'})

    def __len__(self) -> int:
        return len(self.stack)

    def __repr__(self):
        return pformat(self.stack)



from decimal import Decimal

from pytezos.tools.docstring import inline_doc


def format_mutez(value) -> str:
    """ Format amount in mutez (10^-6).

    :param value: can be None (==0), Decimal (treated as tez), int (treated as mutez)
    :returns: amount in mutez, stringified
    """
    if value is None:
        value = 0
    elif isinstance(value, Decimal):
        value = int(value * 10 ** 6)
    elif isinstance(value, float):
        raise ValueError('Please use decimal instead of float')
    return str(value)


def format_tez(value) -> Decimal:
    """ Format amount in tez.

    :param value: can be None (==0), Decimal (treated as tez), int (treated as mutez)
    :rtype: Decimal
    """
    tez = Decimal(format_mutez(value)) / 10 ** 6
    return tez.quantize(Decimal('0.000001'))


class ContentMixin:

    def operation(self, content):
        return content

    @inline_doc
    def endorsement(self, level: int):
        """ Endorse a block.

        :param level: Endorsed level
        :returns: dict or OperationGroup
        """
        return self.operation({
            'kind': 'endorsement',
            'level': str(level)
        })

    @inline_doc
    def seed_nonce_revelation(self, level: int, nonce):
        """ Reveal the nonce committed operation in the previous cycle.
        More info https://tezos.stackexchange.com/questions/567/what-are-nonce-revelations

        :param level: When nonce hash was committed
        :param nonce: Hex string
        :returns: dict or OperationGroup
        """
        return self.operation({
            'kind': 'seed_nonce_revelation',
            'level': str(level),
            'nonce': nonce
        })

    @inline_doc
    def double_endorsement_evidence(self, op1: dict, op2: dict):
        """ Provide evidence of double endorsement (endorsing two different blocks at the same block height).

        :param op1: Inline endorsement
            {
                "branch": $block_hash,
                "operations": {
                    "kind": "endorsement",
                    "level": integer ∈ [-2^31-2, 2^31+2]
                },
                "signature"?: $Signature
            }
        :param op2: Inline endorsement
        :returns: dict or OperationGroup
        """
        return self.operation({
            'kind': 'double_endorsement_evidence',
            'op1': op1,
            'op2': op2
        })

    @inline_doc
    def double_baking_evidence(self, bh1, bh2):
        """ Provide evidence of double baking (two different blocks at the same height).

        :param bh1: First block hash
        :param bh2: Second block hash
        :returns: dict or OperationGroup
        """
        return self.operation({
            'kind': 'double_baking_evidence',
            'bh1': bh1,
            'bh2': bh2
        })

    @inline_doc
    def activate_account(self, activation_code='', pkh=''):
        """ Activate recommended allocations for contributions to the TF fundraiser.
        More info https://activate.tezos.com/

        :param activation_code: Secret code from pdf, leave empty for autocomplete
        :param pkh: Public key hash, leave empty for autocomplete
        :returns: dict or OperationGroup
        """
        return self.operation({
            'kind': 'activate_account',
            'pkh': pkh,
            'secret': activation_code
        })

    @inline_doc
    def proposals(self, proposals,
                  source='', period=0):
        """ Submit and/or upvote proposals to amend the protocol.
        Can only be submitted during a proposal period.
        More info https://tezos.gitlab.io/master/whitedoc/voting.html

        :param proposals: List of proposal hashes or single proposal hash
        :param source: Public key hash (of the signatory), leave None for autocomplete
        :param period: Number of the current voting period, leave 0 for autocomplete
        :returns: dict or OperationGroup
        """
        if not isinstance(proposals, list):
            proposals = [proposals]

        return self.operation({
            'kind': 'proposals',
            'source': source,
            'period': str(period),
            'proposals': proposals
        })

    @inline_doc
    def ballot(self, proposal, ballot,
               source='', period=0):
        """ Vote for a proposal in a given voting period.
        Can only be submitted during Testing_vote or Promotion_vote periods, and only once per period.
        More info https://tezos.gitlab.io/master/whitedoc/voting.html

        :param proposal: Hash of the proposal
        :param ballot: 'Yay', 'Nay' or 'Pass'
        :param source: Public key hash (of the signatory), leave None for autocomplete
        :param period: Number of the current voting period, leave None for autocomplete
        :returns: dict or OperationGroup
        """
        return self.operation({
            'kind': 'ballot',
            'source': source,
            'period': str(period),
            'proposal': proposal,
            'ballot': ballot
        })

    @inline_doc
    def reveal(self, public_key='',
               source='', counter=0, fee=0, gas_limit=0, storage_limit=0):
        """ Reveal the public key associated with a tz address.

        :param public_key: Public key to reveal, Base58 encoded
        :param source: Public key hash of the key revealed, leave None to use signatory address
        :param counter: Current account counter, leave None for autocomplete \
        (More info https://tezos.stackexchange.com/questions/632/how-counter-grows)
        :param fee: Leave None for autocomplete
        :param gas_limit: Leave None for autocomplete
        :param storage_limit: Leave None for autocomplete
        :returns: dict or OperationGroup
        """
        return self.operation({
            'kind': 'reveal',
            'source': source,
            'fee': format_mutez(fee),
            'counter': str(counter),
            'gas_limit': str(gas_limit),
            'storage_limit': str(storage_limit),
            'public_key': public_key
        })

    @inline_doc
    def transaction(self, destination, amount=0, parameters=None,
                    source='', counter=0, fee=0, gas_limit=0, storage_limit=0):
        """ Transfer tez to a given address (implicit or originated).
        If the receiver is a smart contract, then optional parameters may be passed.

        :param source: Address from which funds will be sent, leave None to use signatory address
        :param destination: Address
        :param amount: Amount to send in microtez (int) or tez (Decimal) (optional)
        :param counter: Current account counter, leave None for autocomplete
        :param parameters: { "entrypoint": $string, "value": $Micheline expression } (optional)
        :param fee: Leave None for autocomplete
        :param gas_limit: Leave None for autocomplete
        :param storage_limit: Leave None for autocomplete
        :returns: dict or OperationGroup
        """
        content = {
            'kind': 'transaction',
            'source': source,
            'fee': format_mutez(fee),
            'counter': str(counter),
            'gas_limit': str(gas_limit),
            'storage_limit': str(storage_limit),
            'amount': format_mutez(amount),
            'destination': destination
        }

        if parameters is not None:
            content['parameters'] = parameters

        return self.operation(content)

    @inline_doc
    def origination(self, script, balance=0, delegate=None,
                    source='', counter=0, fee=0, gas_limit=0, storage_limit=0):
        """ Deploy smart contract (scriptless KT accounts are not used for delegation since Babylon)

        :param script: {"code": $Micheline, "storage": $Micheline}
        :param balance: Amount transferred on the balance, WARNING: there is no default way to withdraw funds. \
        More info: https://tezos.stackexchange.com/questions/1315/can-i-withdraw-funds-from-an-empty-smart-contract
        :param delegate: Set contract delegate, default None
        :param source: Address from which funds will be sent, leave None to use signatory address
        :param counter: Current account counter, leave None for autocomplete
        :param fee: Leave None for autocomplete
        :param gas_limit: Leave None for autocomplete
        :param storage_limit: Leave None for autocomplete
        :returns: dict or OperationGroup
        """
        content = {
            'kind': 'origination',
            'source': source,
            'fee': format_mutez(fee),
            'counter': str(counter),
            'gas_limit': str(gas_limit),
            'storage_limit': str(storage_limit),
            'balance': format_mutez(balance),
            'script': script
        }

        if delegate is not None:
            content['delegate'] = delegate

        return self.operation(content)

    @inline_doc
    def delegation(self, delegate='',
                   source='', counter=0, fee=0, gas_limit=0, storage_limit=0):
        """ Delegate funds or register yourself as a delegate.

        :param delegate: tz address of delegate, leave None to register yourself as a delegate
        :param source: Address from which funds will be delegated, leave None to use signatory address
        :param counter: Current account counter, leave None for autocomplete
        :param fee: Leave None for autocomplete
        :param gas_limit: Leave None for autocomplete
        :param storage_limit: Leave None for autocomplete
        :returns: dict or OperationGroup
        """
        return self.operation({
            'kind': 'delegation',
            'source': source,
            'fee': format_mutez(fee),
            'counter': str(counter),
            'gas_limit': str(gas_limit),
            'storage_limit': str(storage_limit),
            'delegate': delegate
        })



from pprint import pformat
from copy import deepcopy

from pytezos.michelson.pack import get_key_hash
from pytezos.michelson.converter import micheline_to_michelson
from pytezos.encoding import is_pkh, is_kt, is_chain_id
from pytezos.repl.parser import parse_expression, parse_prim_expr, assert_expr_equal, assert_comparable, \
    expr_equal, assert_type, get_prim_args, assert_big_map_val, Unit as UnitNone


def assert_stack_item(item: 'StackItem'):
    assert isinstance(item, StackItem), f'expected StackItem, got {type(item).__name__}'


def assert_stack_type(item: 'StackItem', item_type):
    if not isinstance(item_type, list):
        item_type = [item_type]
    expected = ' or '.join(map(lambda x: x.__name__, item_type))
    assert type(item) in item_type, f'expected {expected}, got {type(item).__name__}'


def dispatch_type_map(a, b, mapping):
    assert (type(a), type(b)) in mapping, \
        f'unsupported argument types {type(a).__name__} and {type(b).__name__}'
    return mapping[(type(a), type(b))]


def assert_equal_types(a, b):
    assert type(a) == type(b), f'different types {type(a).__name__} and {type(b).__name__}'


def get_name(annots: list, prefix, default=None):
    if isinstance(annots, list):
        return next((x[1:] for x in annots if x.startswith(prefix)), default)
    return default


class StackItem:
    prim = ''
    __cls__ = {}

    def __init__(self, val, val_expr, type_expr, **kwargs):
        self._val = val
        self.val_expr = val_expr
        self.type_expr = type_expr
        self.name = kwargs.get('name')

    @staticmethod
    def _get_type(type_expr):
        prim, args = parse_prim_expr(type_expr)
        assert prim in StackItem.__cls__, f'{prim}: unknown type primitive'
        item_cls, args_len = StackItem.__cls__[type_expr['prim']]
        assert len(args) == args_len, f'{prim}: expected {args_len} arg(s), got {len(args)}'
        return item_cls

    @classmethod
    def init(cls, val_expr, type_expr, **kwargs):
        return cls(val=parse_expression(val_expr, type_expr),
                   val_expr=val_expr,
                   type_expr=type_expr,
                   **kwargs)

    @staticmethod
    def parse(val_expr, type_expr, **kwargs):
        return StackItem._get_type(type_expr).init(val_expr, type_expr, **kwargs)

    @classmethod
    def __init_subclass__(cls, prim='', args_len=0, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.__cls__[prim] = cls, args_len
        cls.prim = prim

    def __cmp__(self, other: 'StackItem') -> int:
        assert_stack_item(other)
        assert_expr_equal(self.type_expr, other.type_expr)
        if self._val > other._val:
            return 1
        elif self._val < other._val:
            return -1
        else:
            return 0

    def __eq__(self, other: 'StackItem'):
        assert_stack_item(other)
        assert_expr_equal(self.type_expr, other.type_expr)
        return expr_equal(self.val_expr, other.val_expr)

    def __int__(self):
        assert_type(self._val, int)
        return self._val

    def __str__(self):
        assert_type(self._val, str)
        return self._val

    def __bytes__(self):
        assert_type(self._val, bytes)
        return self._val

    def __bool__(self):
        assert_type(self._val, bool)
        return self._val

    def __len__(self):
        return len(self._val)

    def __repr__(self):
        return str(self._val)

    def _modify(self, cache_val=False, **kwargs):
        param = dict(
            val_expr=kwargs.get('val_expr', deepcopy(self.val_expr)),
            type_expr=kwargs.get('type_expr', deepcopy(self.type_expr)),
            name=kwargs.get('name', self.name)
        )
        if 'val' in kwargs or cache_val:
            param['val'] = kwargs.get('val', deepcopy(self._val))
        else:
            param['val'] = parse_expression(param['val_expr'], param['type_expr'])
        return type(self)(**param)

    def __deepcopy__(self, memodict={}):
        return self._modify(cache_val=True)

    def rename(self, annots: list):
        return self._modify(cache_val=True, name=get_name(annots, prefix='@'))


class String(StackItem, prim='string'):

    def __init__(self, val='', val_expr=None, type_expr=None, **kwargs):
        assert isinstance(val, str)
        super(String, self).__init__(val=val if val_expr is None else parse_expression(val_expr, type_expr),
                                     val_expr=val_expr or {'string': val},
                                     type_expr=type_expr or {'prim': self.prim},
                                     **kwargs)

    def __getitem__(self, item):
        assert_type(item, slice)
        return type(self)(self._val[item.start:item.stop])


class Int(StackItem, prim='int'):

    def __init__(self, val=0, val_expr=None, type_expr=None, **kwargs):
        assert_type(val, int)
        assert isinstance(val, int)
        super(Int, self).__init__(val=val if val_expr is None else parse_expression(val_expr, type_expr),
                                  val_expr=val_expr or {'int': str(val)},
                                  type_expr=type_expr or {'prim': self.prim},
                                  **kwargs)


class Bytes(StackItem, prim='bytes'):

    def __init__(self, val=b'', type_expr=None, val_expr=None, **kwargs):
        assert_type(val, bytes)
        super(Bytes, self).__init__(val=val if val_expr is None else parse_expression(val_expr, type_expr),
                                    val_expr=val_expr or {'bytes': val.hex()},
                                    type_expr=type_expr or {'prim': self.prim},
                                    **kwargs)

    def __getitem__(self, item):
        assert_type(item, slice)
        return type(self)(self._val[item.start:item.stop])

    def __repr__(self):
        return self._val.hex()


class Nat(Int, prim='nat'):

    def __init__(self, val=0, val_expr=None, type_expr=None, **kwargs):
        assert_type(val, int)
        assert val >= 0, 'expected non-negative val'
        super(Nat, self).__init__(val=val, val_expr=val_expr, type_expr=type_expr, **kwargs)


class Bool(StackItem, prim='bool'):

    def __init__(self, val=False, val_expr=None, type_expr=None, **kwargs):
        assert_type(val, bool)
        super(Bool, self).__init__(val=val if val_expr is None else parse_expression(val_expr, type_expr),
                                   val_expr=val_expr or {'prim': str(val)},
                                   type_expr=type_expr or {'prim': self.prim},
                                   **kwargs)


class Unit(StackItem, prim='unit'):

    def __init__(self, val=UnitNone(), val_expr=None, type_expr=None, **kwargs):
        super(Unit, self).__init__(val=val if val_expr is None else parse_expression(val_expr, type_expr),
                                   val_expr=val_expr or {'prim': 'Unit'},
                                   type_expr=type_expr or {'prim': self.prim},
                                   **kwargs)


class List(StackItem, prim='list', args_len=1):

    def __iter__(self):
        for item in self.val_expr:
            yield self.parse(val_expr=item, type_expr=self.type_expr['args'][0])

    def __getitem__(self, item):
        assert_type(item, int)
        return self.parse(val_expr=self.val_expr[item], type_expr=self.type_expr['args'][0])

    def prepend(self, item: StackItem) -> 'List':
        self.assert_val_type(item)
        return self._modify(val_expr=[item.val_expr] + self.val_expr)

    def cut_head(self):
        assert len(self._val) > 0, f'must be non-empty list'
        head = self.parse(val_expr=self.val_expr[0], type_expr=self.type_expr['args'][0])
        tail = self._modify(val_expr=self.val_expr[1:])
        return head, tail

    @classmethod
    def empty(cls, v_type_expr):
        return cls(val=[], val_expr=[], type_expr={'prim': cls.prim, 'args': [v_type_expr]})

    @classmethod
    def new(cls, items: list):
        assert isinstance(items, list) and len(items) > 0, f'expected non-empty list'
        return cls.init(val_expr=[assert_stack_item(x) or x.val_expr for x in items],
                        type_expr={'prim': cls.prim, 'args': [items[0].type_expr]})

    def assert_val_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][0], item.type_expr)

    def val_type(self):
        return self._get_type(self.type_expr['args'][0])


class Pair(StackItem, prim='pair', args_len=2):

    @classmethod
    def new(cls, left: StackItem, right: StackItem):
        assert_stack_item(left)
        assert_stack_item(right)
        return cls.init(val_expr={'prim': 'Pair', 'args': [left.val_expr, right.val_expr]},
                        type_expr={'prim': cls.prim, 'args': [left.type_expr, right.type_expr]})

    def get_element(self, idx: int):
        assert idx in [0, 1], f'unexpected element index {idx}'
        type_expr = self.type_expr['args'][idx]
        inferred_name = get_name(type_expr.get('annots'), '%', default=['car', 'cdr'][idx])
        if self.name:
            inferred_name = f'{self.name}.{inferred_name}'

        return self.parse(val_expr=self.val_expr['args'][idx],
                          type_expr=type_expr,
                          name=inferred_name)


class Option(StackItem, prim='option', args_len=1):

    @classmethod
    def some(cls, item: StackItem):
        assert_stack_item(item)
        return cls.init(val_expr={'prim': 'Some', 'args': [item.val_expr]},
                        type_expr={'prim': cls.prim, 'args': [item.type_expr]})

    @classmethod
    def none(cls, type_expr):
        return cls(val=None,
                   val_expr={'prim': 'None'},
                   type_expr={'prim': cls.prim, 'args': [type_expr]})

    def is_none(self):
        return self._val is None

    def get_some(self):
        return self.parse(val_expr=self.val_expr['args'][0],
                          type_expr=self.type_expr['args'][0])


class Or(StackItem, prim='or', args_len=2):

    @classmethod
    def left(cls, r_type_expr, item: StackItem):
        assert_stack_item(item)
        return cls.init(val_expr={'prim': 'Left', 'args': [item.val_expr]},
                        type_expr={'prim': cls.prim, 'args': [item.type_expr, r_type_expr]})

    @classmethod
    def right(cls, l_type_expr, item: StackItem):
        assert_stack_item(item)
        return cls.init(val_expr={'prim': 'Right', 'args': [item.val_expr]},
                        type_expr={'prim': cls.prim, 'args': [l_type_expr, item.type_expr]})

    def is_left(self):
        prim, _ = parse_prim_expr(self.val_expr)
        return prim == 'Left'

    def get_some(self):
        idx = 0 if self.is_left() else 1
        return self.parse(val_expr=self.val_expr['args'][0],
                          type_expr=self.type_expr['args'][idx])


class Set(StackItem, prim='set', args_len=1):

    def __iter__(self):
        for item in self.val_expr:
            yield self.parse(val_expr=item, type_expr=self.type_expr['args'][0])

    @classmethod
    def empty(cls, k_type_expr):
        assert_comparable(k_type_expr)
        return cls(val=[], val_expr=[], type_expr={'prim': cls.prim, 'args': [k_type_expr]})

    @classmethod
    def new(cls, items: list):
        assert isinstance(items, list) and len(items) > 0, f'expected non-empty list'
        assert len(set(items)) == len(items), f'duplicate keys found'
        return cls.init(val_expr=[assert_stack_item(x) or x.val_expr for x in items],
                        type_expr={'prim': cls.prim, 'args': [items[0].type_expr]})

    def __contains__(self, item: StackItem):
        self.assert_key_type(item)
        return item._val in self._val

    def add(self, item: StackItem) -> 'Set':
        self.assert_key_type(item)
        if item._val in self._val:
            return self
        return self._modify(val_expr=[item.val_expr] + self.val_expr)  # TODO: sort

    def remove(self, item: StackItem) -> 'Set':
        self.assert_key_type(item)
        if item._val not in self._val:
            return self
        return self._modify(val_expr=[x for x in self.val_expr if not expr_equal(x, item.val_expr)])

    def assert_key_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][0], item.type_expr)


class Map(StackItem, prim='map', args_len=2):

    def __iter__(self):
        for elt in self.val_expr:
            yield tuple(self.parse(val_expr=elt['args'][i], type_expr=self.type_expr['args'][i]) for i in [0, 1])

    @classmethod
    def empty(cls, k_type_expr, v_type_expr):
        assert_comparable(k_type_expr)
        return cls(val={}, val_expr=[], type_expr={'prim': cls.prim, 'args': [k_type_expr, v_type_expr]})

    @classmethod
    def new(cls, items: list):
        assert isinstance(items, list) and len(items) > 0, f'expected non-empty list'
        _ = [assert_stack_item(x) for kv in items for x in kv]
        k0, v0 = items[0]
        return cls.init(val_expr=[{'prim': 'Elt', 'args': [k.val_expr, v.val_expr]} for k, v in items],
                        type_expr={'prim': cls.prim, 'args': [k0.type_expr, v0.type_expr]})

    def __contains__(self, key: StackItem):
        self.assert_key_type(key)
        return key._val in self._val

    def update(self, key: StackItem, val: StackItem) -> 'Map':
        self.assert_key_type(key)
        self.assert_val_type(val)
        upd = {'prim': 'Elt', 'args': [key.val_expr, val.val_expr]}
        if key._val in self._val:
            val_expr = [upd if expr_equal(x['args'][0], key.val_expr) else x for x in self.val_expr]
        else:
            val_expr = self.val_expr + [upd]  # TODO: sort
        return self._modify(val_expr=val_expr)

    def remove(self, key: StackItem) -> 'Map':
        self.assert_key_type(key)
        if key._val not in self._val:
            return self
        val_expr = [x for x in self.val_expr if not expr_equal(x['args'][0], key.val_expr)]
        return self._modify(val_expr=val_expr)

    def find(self, key: StackItem) -> 'StackItem':
        self.assert_key_type(key)
        if key._val in self._val:
            val_expr = next(elt['args'][1] for elt in self.val_expr if expr_equal(elt['args'][0], key.val_expr))
            return self.parse(val_expr=val_expr, type_expr=self.type_expr['args'][1])

    def assert_key_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][0], item.type_expr)

    def assert_val_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][1], item.type_expr)

    def val_type_expr(self):
        return self.type_expr['args'][1]

    def make_key(self, k_val_expr):
        return self.parse(k_val_expr, self.type_expr['args'][0])

    def make_val(self, v_val_expr):
        return self.parse(v_val_expr, self.type_expr['args'][1])


class BigMap(StackItem, prim='big_map', args_len=2):

    def _get_key_hash(self, key: StackItem):
        self.assert_key_type(key)
        return get_key_hash(key.val_expr, self.type_expr['args'][0])

    def __contains__(self, key: StackItem):
        return self._get_key_hash(key) in self.val_expr['_diff']

    def find(self, key: StackItem) -> 'StackItem':
        key_hash = self._get_key_hash(key)
        v_val_expr = self.val_expr['_diff'].get(key_hash, {}).get('value')
        if v_val_expr is not None:
            return self.parse(val_expr=v_val_expr, type_expr=self.type_expr['args'][1])

    def _update(self, key: StackItem, value: StackItem = None) -> 'BigMap':
        key_hash = self._get_key_hash(key)
        update = {'action': 'update',
                  'big_map': str(self._val),
                  'key_hash': key_hash,
                  'key': key.val_expr}
        if value is not None:
            self.assert_val_type(value)
            update['value'] = value.val_expr

        val_expr = deepcopy(self.val_expr)
        val_expr['_diff'][key_hash] = update
        return self._modify(val_expr=val_expr)

    def update(self, key: StackItem, value: StackItem) -> 'BigMap':
        return self._update(key, value)

    def remove(self, key: StackItem) -> 'BigMap':
        return self._update(key)

    def assert_key_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][0], item.type_expr)

    def assert_val_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][1], item.type_expr)

    def val_type_expr(self):
        return self.type_expr['args'][1]


class Timestamp(Int, prim='timestamp'):
    pass


class Mutez(Nat, prim='mutez'):

    def __init__(self, val=0, val_expr=None, type_expr=None, **kwargs):
        assert_type(val, int)
        assert val.bit_length() <= 63, f'mutez overflow, got {val.bit_length()} bits, should not exceed 63'
        super(Mutez, self).__init__(val=val, val_expr=val_expr, type_expr=type_expr, **kwargs)


class Address(String, prim='address'):

    @classmethod
    def new(cls, address):
        assert is_pkh(address) or is_kt(address), f'expected address, got {address}'
        return cls(val=address,
                   val_expr={'string': address},
                   type_expr={'prim': cls.prim})


class Contract(StackItem, prim='contract', args_len=1):

    @classmethod
    def new(cls, contract, type_expr):
        assert is_pkh(contract[:36]) or is_kt(contract[:36]), f'expected contract, got {contract}'
        if len(contract) > 36:
            assert contract[36] == '%', f'expected contract, got {contract}'
        return cls(val=contract,
                   val_expr={'string': contract},
                   type_expr={'prim': cls.prim, 'args': [type_expr]})

    def assert_param_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][0], item.type_expr)

    def get_address(self):
        return self._val[:36]

    def get_entrypoint(self):
        annot = self._val[36:] or '%default'
        return annot[1:]


class Key(StackItem, prim='key'):
    pass


class KeyHash(String, prim='key_hash'):

    @classmethod
    def new(cls, key_hash):
        assert is_pkh(key_hash), f'expected key hash, got {key_hash}'
        return cls(val=key_hash,
                   val_expr={'string': key_hash},
                   type_expr={'prim': cls.prim})


class Signature(StackItem, prim='signature'):
    pass


class ChainID(String, prim='chain_id'):

    def __init__(self, val, val_expr=None, type_expr=None, **kwargs):
        assert is_chain_id(val), f'expected chain_id, got {val}'
        super(ChainID, self).__init__(val, val_expr=val_expr, type_expr=type_expr, **kwargs)


class Lambda(StackItem, prim='lambda', args_len=2):

    @classmethod
    def new(cls, p_type_expr, r_type_expr, code):
        return cls(val=code,
                   val_expr=code,
                   type_expr={'prim': 'lambda', 'args': [p_type_expr, r_type_expr]})

    def assert_param_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][0], item.type_expr)

    def assert_ret_type(self, item: StackItem):
        assert_stack_item(item)
        assert_expr_equal(self.type_expr['args'][1], item.type_expr)

    @property
    def code(self):
        return self._val

    def partial_apply(self, item: StackItem):
        l_type_expr, r_type_expr = get_prim_args(self.type_expr['args'][0], prim='pair', args_len=2)
        assert_equal_types(l_type_expr, item.type_expr)
        code = [{'prim': 'PUSH', 'args': [item.type_expr, item.val_expr]}, {'prim': 'PAIR'}] + self.val_expr
        return self._modify(val=code,
                            val_expr=code,
                            type_expr={'prim': 'lambda', 'args': [r_type_expr, self.type_expr['args'][1]]})

    def __repr__(self):
        return micheline_to_michelson(self._val, inline=True)


class Operation(StackItem, prim='operation'):

    def __init__(self, val, val_expr, type_expr, **kwargs):
        super(Operation, self).__init__(val, val_expr, type_expr, **kwargs)

    @staticmethod
    def format_content(content):
        if content['kind'] == 'transaction':
            if content['destination'].startswith('tz'):
                return f'<send {content["amount"]} to {content["destination"]}>'
            else:
                return f'<call {content["destination"]}%{content["parameters"]["entrypoint"]}>'
        elif content['kind'] == 'origination':
            return f'<originate {content["originated_contract"]}>'
        elif content['kind'] == 'delegation':
            if content.get('delegate'):
                return f'<delegate to {content["delegate"]}>'
            else:
                return f'<unset delegate>'
        else:
            assert False, f'unexpected content {pformat(content, compact=True)}'

    @classmethod
    def new(cls, content):
        return cls(val=cls.format_content(content),
                   val_expr={'string': cls.format_content(content), '_content': content},
                   type_expr={'prim': cls.prim})

    @property
    def content(self):
        return self.val_expr.get('_content')



from typing import Dict
from simplejson import JSONDecodeError

from pytezos.interop import Interop
from pytezos.repl.types import StackItem, Map, BigMap
from pytezos.repl.parser import parse_expression, assert_expr_equal, get_int, assert_comparable, assert_big_map_val
from pytezos.michelson.pack import get_key_hash


def make_elt(args):
    assert isinstance(args, list) and len(args) == 2
    return {'prim': 'Elt', 'args': args}


def elt_to_update(elt, type_expr, big_map_id):
    key_hash = get_key_hash(elt['args'][0], type_expr['args'][0])
    update = {'action': 'update',
              'big_map': str(big_map_id),
              'key_hash': key_hash,
              'key': elt['args'][0],
              'value': elt['args'][1]}
    return key_hash, update


class BigMapPool:

    def __init__(self):
        self.maps = {}  # type: Dict[int, Map]
        self.tmp_id = -1
        self.alloc_id = 0
        self.maybe_remove = set()

    def reset(self):
        self.maps.clear()
        self.tmp_id = -1
        self.alloc_id = 0
        self.maybe_remove.clear()

    def empty(self, k_type_expr, v_type_expr) -> BigMap:
        assert_comparable(k_type_expr)
        assert_big_map_val(v_type_expr)
        res = BigMap(val=self.tmp_id,
                     val_expr={'int': str(self.tmp_id), '_diff': {}},
                     type_expr={'prim': 'big_map', 'args': [k_type_expr, v_type_expr]})
        self.tmp_id -= 1
        return res

    def _pre_alloc(self, val_expr, type_expr):
        res = {'int': str(self.tmp_id),
               '_diff': dict(elt_to_update(elt, type_expr, self.tmp_id) for elt in val_expr)}
        self.tmp_id -= 1
        return res

    def _check_allocated(self, val_expr, type_expr, network=None):
        big_map_id = get_int(val_expr)
        assert big_map_id >= 0, f'expected an allocated big map (>=0), got {big_map_id}'
        if big_map_id in self.maps:
            big_map = self.maps[big_map_id]
            _ = [assert_expr_equal(type_expr['args'][i], big_map.type_expr['args'][i]) for i in [0, 1]]
            return big_map_id, {'_diff': {}}
        else:
            assert network, f'big map #{big_map_id} is not allocated'
            self.alloc_id = max(self.alloc_id, big_map_id + 1)
            return big_map_id, {'_diff': {}, '_network': network}

    def _pre_copy(self, val_expr, type_expr, network=None):
        big_map_id, kw = self._check_allocated(val_expr, type_expr, network)
        res = {'int': str(self.tmp_id), '_copy': big_map_id, **kw}
        self.tmp_id -= 1
        return res

    def _pre_remove(self, val_expr, type_expr, network=None):
        big_map_id, kw = self._check_allocated(val_expr, type_expr, network)
        self.maybe_remove.add(big_map_id)
        return {'int': str(big_map_id), **kw}

    def pre_alloc(self, val_expr, type_expr, copy=False, network=None):
        def alloc_selector(val_node, type_node, res, type_path):
            prim = type_node['prim']
            if prim in ['list', 'set']:
                return res
            if prim in ['pair', 'or']:
                return {'prim': val_node['prim'], 'args': res}
            elif prim == 'option' and val_node['prim'] == 'Some':
                return {'prim': val_node['prim'], 'args': res}
            elif prim == 'map':
                return list(map(make_elt, res))
            elif prim == 'big_map':
                if isinstance(val_node, list):
                    return self._pre_alloc(val_node, type_node)
                elif copy:
                    return self._pre_copy(val_node, type_node, network=network)
                else:
                    return self._pre_remove(val_node, type_node, network=network)

            return val_node

        val_expr = parse_expression(val_expr, type_expr, alloc_selector)
        return StackItem.parse(val_expr=val_expr, type_expr=type_expr)

    def diff(self, storage: StackItem):
        res = []
        alloc_id = self.alloc_id
        maybe_remove = self.maybe_remove

        def diff_selector(val_node, type_node, val, type_path):
            nonlocal res, alloc_id, maybe_remove
            prim = type_node['prim']
            if prim in ['list', 'set']:
                return val
            if prim in ['pair', 'or']:
                return {'prim': val_node['prim'], 'args': val}
            elif prim == 'option' and val_node['prim'] == 'Some':
                return {'prim': val_node['prim'], 'args': val}
            elif prim == 'map':
                return list(map(make_elt, val))
            elif prim == 'big_map':
                assert isinstance(val, int), f'expected big map pointer'
                if val < 0:
                    big_map_id = alloc_id
                    if val_node.get('_copy'):
                        res.append({'action': 'copy',
                                    'source_big_map': str(val_node['_copy']),
                                    'destination_big_map': str(big_map_id)})
                    else:
                        res.append({'action': 'alloc',
                                    'big_map': str(big_map_id),
                                    'key_type': type_node['args'][0],
                                    'value_type': type_node['args'][1]})
                    alloc_id += 1
                else:
                    big_map_id = val
                    maybe_remove.remove(big_map_id)

                res.extend(map(lambda x: {**x, 'big_map': str(big_map_id)}, val_node['_diff'].values()))
                return {'int': str(big_map_id)}
            else:
                return val_node

        val_expr = parse_expression(storage.val_expr, storage.type_expr, diff_selector)
        res.extend([{'action': 'remove', 'big_map': x} for x in maybe_remove])
        return StackItem.parse(val_expr, storage.type_expr), res

    def commit(self, big_map_diff: list):
        for diff in big_map_diff:
            big_map_id = int(diff['big_map'])
            raw_map = self.maps.get(big_map_id)
            if diff['action'] == 'alloc':
                assert raw_map is None, f'#{big_map_id} already allocated'
                self.maps[big_map_id] = Map.empty(k_type_expr=diff['key_type'], v_type_expr=diff['value_type'])
            elif diff['action'] == 'remove':
                assert raw_map is not None, f'#{big_map_id} is not allocated'
                del self.maps[big_map_id]
            elif diff['action'] == 'update':
                assert raw_map is not None, f'#{big_map_id} is not allocated'
                key = raw_map.make_key(diff['key'])
                if diff.get('value'):
                    val = raw_map.make_val(diff['value'])
                    self.maps[big_map_id] = raw_map.update(key, val)
                else:
                    self.maps[big_map_id] = raw_map.remove(key)

        if big_map_diff:
            self.alloc_id = max([int(x['big_map']) for x in big_map_diff]) + 1
        self.maybe_remove.clear()

    def _get_big_map_val(self, big_map: BigMap, key: StackItem):
        key_hash = get_key_hash(key.val_expr, key.type_expr)
        network = big_map.val_expr['_network']
        try:
            res = Interop().using(network).shell.head.context.big_maps[int(big_map)][key_hash]()
        except JSONDecodeError:
            res = None
        return res

    def contains(self, big_map: BigMap, key: StackItem) -> bool:
        if key in big_map:
            return big_map.find(key) is not None
        if int(big_map) >= 0:
            if big_map.val_expr.get('_network'):
                v_val_expr = self._get_big_map_val(big_map, key)
                return v_val_expr is not None
            else:
                assert int(big_map) in self.maps, f'#{int(big_map)} is not allocated'
                return key in self.maps[int(big_map)]
        return False

    def find(self, big_map: BigMap, key: StackItem) -> 'StackItem':
        if key in big_map:
            return big_map.find(key)
        if int(big_map) >= 0:
            if big_map.val_expr.get('_network'):
                v_val_expr = self._get_big_map_val(big_map, key)
                if v_val_expr:
                    return StackItem.parse(val_expr=v_val_expr, type_expr=big_map.type_expr['args'][1])
            else:
                assert int(big_map) in self.maps, f'#{int(big_map)} is not allocated'
                return self.maps[int(big_map)].find(key)



from pytezos.rpc.shell import *
from pytezos.rpc.protocol import *
from pytezos.rpc.helpers import *
from pytezos.rpc.search import *
from pytezos.rpc.node import RpcNode, RpcMultiNode


class RpcProvider:

    def __init__(self, klass=RpcNode, **urls):
        self.urls = urls
        self.klass = klass

    @lru_cache(maxsize=None)
    def __getattr__(self, network) -> ShellQuery:
        return ShellQuery(node=self.klass(uri=self.urls[network], network=network))

    def __dir__(self):
        return list(super(RpcProvider, self).__dir__()) + list(self.urls.keys())

    def __repr__(self):
        res = [
            super(RpcProvider, self).__repr__(),
            '\nNetworks',
            *list(map(lambda x: f'.{x[0]}  # {x[1]}', self.urls.items()))
        ]
        return '\n'.join(res)


localhost = RpcProvider(
    sandboxnet='http://127.0.0.1:8732/',
    bbbox='http://flextesa:20000'
)
tzkt = RpcProvider(
    mainnet='https://rpc.tzkt.io/mainnet/',
    babylonnet='https://rpc.tzkt.io/babylonnet/',
    carthagenet='https://rpc.tzkt.io/carthagenet/',
    zeronet='https://rpc.tzkt.io/zeronet/'
)
giganode = RpcProvider(
    mainnet='https://mainnet-tezos.giganode.io/',
    carthagenet='https://testnet-tezos.giganode.io/',
    labnet='https://labnet-tezos.giganode.io/'
)
pool = RpcProvider(
    klass=RpcMultiNode,
    mainnet=[
        'https://rpc.tzkt.io/mainnet/',
        'https://tezos-prod.cryptonomic-infra.tech/',
        'https://rpc.tezrpc.me/',
        'https://api.tezos.org.ua/',
        'https://api.tez.ie/',

    ],
    babylonnet=[
        'https://rpc.tzkt.io/babylonnet/',
        'https://tezos-dev.cryptonomic-infra.tech/',
    ],
    carthagenet=[
        'https://rpc.tzkt.io/carthagenet/',
        'https://carthagenet.tezos.org.ua/',
    ]
)

mainnet = tzkt.mainnet
babylonnet = tzkt.babylonnet
carthagenet = tzkt.carthagenet
zeronet = tzkt.zeronet
labnet = giganode.labnet



import requests
import simplejson as json
from functools import lru_cache
from binascii import hexlify
from datetime import datetime
from time import sleep

from pytezos.encoding import base58_decode
from pytezos.rpc.query import RpcQuery
from pytezos.tools.docstring import get_attr_docstring
from pytezos.rpc.search import CyclesQuery, VotingPeriodsQuery


def make_operation_result(**kwargs):
    return {'metadata': {'operation_result': kwargs}}


class ShellQuery(RpcQuery, path=''):

    @property
    def blocks(self):
        """ Shortcut for `chains.main.blocks`
        """
        return self.chains.main.blocks

    @property
    def head(self):
        """ Shortcut for `blocks.head`
        """
        return self.blocks.head

    @property
    @lru_cache(maxsize=None)
    def block(self):
        """ Cached head block, useful if you just want to explore things.
        """
        return self.blocks[self.head.hash()]

    @property
    def cycles(self):
        """ Operate on cycles rather than blocks.
        """
        return CyclesQuery(
            node=self.node,
            path=self._wild_path + '/chains/{}/blocks',
            params=self._params + ['main']
        )

    @property
    def voting_periods(self):
        """
        Operate on voting periods rather than blocks.
        """
        return VotingPeriodsQuery(
            node=self.node,
            path=self._wild_path + '/chains/{}/blocks',
            params=self._params + ['main']
        )

    @property
    def contracts(self):
        """ Shortcut for `head.context.contracts`
        """
        return self.head.context.contracts

    @property
    def mempool(self):
        """ Shortcut for `chains.main.mempool`
        """
        return self.chains.main.mempool

    def wait_next_block(self, block_hash=None):
        """ Wait until next block is finalized.

        :param block_hash: Current block hash (optional). If not set, current head is used.
        """
        block_time = int(self.block.context.constants()["time_between_blocks"][0])
        header = self.head.header()
        if block_hash is None:
            block_hash = header['hash']

        prev_block_dt = datetime.strptime(header['timestamp'], '%Y-%m-%dT%H:%M:%SZ')
        elapsed_sec = (datetime.utcnow() - prev_block_dt).seconds
        delay_sec = 0 if elapsed_sec > block_time else block_time - elapsed_sec
        print(f'Wait {delay_sec} seconds until block {block_hash} is finalized')
        sleep(delay_sec)

        for i in range(block_time):
            current_block_hash = self.head.hash()
            if current_block_hash == block_hash:
                sleep(1)
            else:
                return current_block_hash
        assert False


class ChainQuery(RpcQuery, path='/chains/{}'):

    def watermark(self):
        """ Chain watermark, hex encoded.
        """
        data = self.chain_id()
        return hexlify(base58_decode(data.encode())).decode()


class InvalidBlockQuery(RpcQuery, path='/chains/{}/invalid_blocks/{}'):

    def delete(self):
        return self._delete()


class MempoolQuery(RpcQuery, path='/chains/{}/mempool'):

    def post(self, configuration):
        """ Set operation filter rules.

        :param configuration: a JSON dictionary, known keys are `minimal_fees`, `minimal_nanotez_per_gas_unit`, \
        `minimal_nanotez_per_byte`
        """
        return self._post(json=configuration)


class PendingOperationsQuery(RpcQuery, path='/chains/{}/mempool/pending_operations'):

    def __getitem__(self, item):
        """ Search for operation in node's mempool by hash.

        :param item: operation group hash (base58)
        """
        operations_dict = self()
        for status, operations in operations_dict.items():
            for operation in operations:
                if isinstance(operation, dict):
                    if operation['hash'] == item:
                        return {**make_operation_result(status=status), **operation}
                elif isinstance(operation, list):
                    if operation[0] == item:
                        errors = operation[1].pop1('error', default=[])
                        return {**make_operation_result(status=status, errors=errors), **operation[1]}
                else:
                    assert False, operation
        raise StopIteration

    def __repr__(self):
        res = [
            super(PendingOperationsQuery, self).__repr__(),
            '[]' + get_attr_docstring(self.__class__, '__getitem__')
        ]
        return '\n'.join(res)


class DescribeQuery(RpcQuery, path='/describe'):

    def __call__(self, recurse=True):
        """ Get RPCs documentation and input/output schema.

        :param recurse: Show information for child elements, default is True.
        In some cases doesn't work without this flag.
        """
        return super(DescribeQuery, self).__call__(recurse=recurse)

    def __repr__(self):
        res = [
            super(DescribeQuery, self).__repr__(),
            f'(){get_attr_docstring(DescribeQuery, "__call__")}',
            'Can be followed by any path:\n.chains\n.network.connections\netc\n'
        ]
        return '\n'.join(res)


class BlockInjectionQuery(RpcQuery, path='/injection/block'):

    def post(self, block, _async=False, force=False, chain=None):
        """ Inject a block in the node and broadcast it.
        The `operations` embedded in `blockHeader` might be pre-validated using a contextual RPCs from the latest block
        (e.g. '/blocks/head/context/preapply').

        :param block: Json input
            {
                "data": <hex-encoded block header>,
                "operations": [ [ {
                "branch": <block_hash>,
                "data": <hex-encoded operation>
                } ... ] ... ]
            }
        :param _async: By default, the RPC will wait for the block to be validated before answering, \
        set True if you don't want to.
        :param force:
        :param chain: Optionally you can specify the chain
        :returns: ID of the block
        """
        return self._post(
            params={
                'async': _async,
                'force': force,
                'chain': chain
            },
            json=block
        )


class OperationInjectionQuery(RpcQuery, path='/injection/operation'):

    def post(self, operation, _async=False, chain=None):
        """ Inject an operation in node and broadcast it.
        The `signedOperationContents` should be constructed using a contextual RPCs from the latest block
        and signed by the client.

        :param operation: Hex-encoded operation data or bytes
        :param _async: By default, the RPC will wait for the operation to be (pre-)validated before answering, \
        set True if you don't want to.
        :param chain: Optionally you can specify the chain
        :returns: ID of the operation
        """
        if isinstance(operation, bytes):
            operation = operation.hex()

        return self._post(
            params={
                'async': _async,
                'chain': chain
            },
            json=operation
        )


class ProtocolInjectionQuery(RpcQuery, path='/injection/protocol'):

    def post(self, protocol, _async=False, force=False):
        """ Inject a protocol in node.

        :param protocol: Json input
            {
                "expected_env_version": <integer>,
                "components": [{
                    "name": <unistring>,
                    "interface"?: <hex-encoded data>,
                    "implementation": <hex-encoded data> }
                     ...
                ]}
            }
        :param _async:
        :param force:
        :returns: ID of the protocol
        """
        return self._post(
            params={
                'async': _async,
                'force': force
            },
            json=protocol
        )


class ResponseGenerator:

    def __init__(self, res: requests.Response):
        self._lines = res.iter_lines()

    def __iter__(self):
        for line in self._lines:
            yield json.loads(line.decode())


class MonitorQuery(RpcQuery, path=['/monitor/active_chains',
                                   '/monitor/bootstrapped',
                                   '/monitor/commit_hash',
                                   '/monitor/heads/{}',
                                   '/monitor/protocols',
                                   '/monitor/valid_blocks']):

    def __call__(self, *args, **kwargs):
        return ResponseGenerator(self.node.request(
            method='GET',
            path=self.path,
            params=kwargs,
            stream=True
        ))

    def __repr__(self):
        res = [
            super(MonitorQuery, self).__repr__(),
            'NOTE: Returned object is a generator.'
        ]
        return '\n'.join(res)


class ConnectionQuery(RpcQuery, path='/network/connections/{}'):

    def delete(self, wait=False):
        return self._delete(params=dict(wait=wait))


class NetworkItems(RpcQuery, path=['/network/peers', '/network/points']):

    def __call__(self, _filter=None):
        return self._get(params={'filter': _filter})


class NetworkLogQuery(RpcQuery, path=['/network/peers/{}/log', '/network/points/{}/log']):

    def __call__(self, monitor=False):
        if monitor:
            return ResponseGenerator(self.node.request(
                method='GET',
                path=self.path,
                stream=True
            ))
        else:
            return self._get()



from glob import glob
from os.path import abspath, dirname, join
from pprint import pprint
import fire

from pytezos import pytezos, Contract, RpcError
from pytezos.michelson.docstring import generate_docstring
from pytezos.operation.result import OperationResult
from pytezos.tools.github import create_deployment, create_deployment_status

kernel_js_path = join(dirname(dirname(__file__)), 'assets', 'kernel.js')
kernel_json = {
    "argv": ['pytezos', 'kernel', 'run', "-file", "{connection_file}"],
    "display_name": "Michelson",
    "language": "michelson",
    "codemirror_mode": "michelson"
}


def make_bcd_link(network, address):
    net = {
        'carthagenet': 'carthage',
        'babylonnet': 'babylon',
        'sandboxnet': 'sandbox',
        'mainnet': 'main',
        'zeronet': 'zeronet'
    }
    return f'https://better-call.dev/{net[network]}/{address}'


def get_contract(path):
    if path is None:
        files = glob('*.tz')
        assert len(files) == 1
        contract = Contract.from_file(abspath(files[0]))

    elif any(map(lambda x: path.startswith(x), ['zeronet', 'babylonnet', 'mainnet', 'carthagenet'])):
        network, address = path.split(':')
        ptz = pytezos.using(shell=network)
        script = ptz.shell.contracts[address].script()
        contract = Contract.from_micheline(script['code'])

    else:
        contract = Contract.from_file(path)

    return contract


class PyTezosCli:

    def storage(self, action, path=None):
        """ Manage contract storage.

        :param action: One of `schema`, `default`
        :param path: Path to the .tz file, or the following uri: <network>:<KT-address>
        """
        contract = get_contract(path)
        if action == 'schema':
            print(generate_docstring(contract.storage.schema, title='storage'))
        elif action == 'default':
            pprint(contract.storage.default())
        else:
            assert False, action

    def parameter(self, action, path=None):
        """ Manage contract parameter.

        :param action: One of `schema`
        :param path: Path to the .tz file, or the following uri: <network>:<KT-address>
        """
        contract = get_contract(path)
        if action == 'schema':
            print(generate_docstring(contract.parameter.schema, title='parameter'))
        else:
            assert False, action

    def activate(self, path, network='carthagenet'):
        """ Activates and reveals key from the faucet file.

        :param path: Path to the .json file downloaded from https://faucet.tzalpha.net/
        :param network: Default is Babylonnet
        """
        ptz = pytezos.using(key=path, shell=network)
        print(f'Activating {ptz.key.public_key_hash()} in the {network}')

        if ptz.balance() == 0:
            try:
                opg = ptz.activate_account().autofill().sign()
                print(f'Injecting activation operation:')
                pprint(opg.json_payload())
                opg.inject(_async=False)
            except RpcError as e:
                pprint(e)
                exit(-1)
            else:
                print(f'Activation succeeded! Claimed balance: {ptz.balance()} ꜩ')
        else:
            print('Already activated')

        try:
            opg = ptz.reveal().autofill().sign()
            print(f'Injecting reveal operation:')
            pprint(opg.json_payload())
            opg.inject(_async=False)
        except RpcError as e:
            pprint(e)
            exit(-1)
        else:
            print(f'Your key {ptz.key.public_key_hash()} is now active and revealed')

    def deploy(self, path, storage=None, network='carthagenet', key=None,
               github_repo_slug=None, github_oauth_token=None, dry_run=False):
        """ Deploy contract to the specified network.

        :param path: Path to the .tz file
        :param storage: Storage in JSON format (not Micheline)
        :param network:
        :param key:
        :param github_repo_slug:
        :param github_oauth_token:
        :param dry_run: Set this flag if you just want to see what would happen
        """
        ptz = pytezos.using(shell=network, key=key)
        print(f'Deploying contract using {ptz.key.public_key_hash()} in the {network}')

        contract = get_contract(path)
        if storage is not None:
            storage = contract.storage.encode(storage)

        try:
            opg = ptz.origination(script=contract.script(storage=storage)).autofill().sign()
            print(f'Injecting origination operation:')
            pprint(opg.json_payload())

            if dry_run:
                pprint(opg.preapply())
                exit(0)
            else:
                opg = opg.inject(_async=False)
        except RpcError as e:
            pprint(e)
            exit(-1)
        else:
            originated_contracts = OperationResult.originated_contracts(opg)
            assert len(originated_contracts) == 1
            bcd_link = make_bcd_link(network, originated_contracts[0])
            print(f'Contract was successfully deployed: {bcd_link}')

            if github_repo_slug:
                deployment = create_deployment(github_repo_slug, github_oauth_token,
                                               environment=network)
                pprint(deployment)
                status = create_deployment_status(github_repo_slug, github_oauth_token,
                                                  deployment_id=deployment['id'],
                                                  state='success',
                                                  environment=network,
                                                  environment_url=bcd_link)
                pprint(status)


def main():
    return fire.Fire(PyTezosCli)


if __name__ == '__main__':
    main()



import io
import os
import tarfile
import netstruct
import requests
import simplejson as json
from tqdm import tqdm
from binascii import hexlify
from collections import OrderedDict
from tempfile import TemporaryDirectory
from typing import List, Tuple

from pytezos.crypto import blake2b_32
from pytezos.encoding import base58_encode
from pytezos.tools.diff import make_patch, apply_patch, generate_unidiff_html
from pytezos.tools.docstring import get_class_docstring, InlineDocstring


def dir_to_files(path) -> List[Tuple[str, str]]:
    files = list()

    with open(os.path.join(path, 'TEZOS_PROTOCOL')) as f:
        index = json.load(f)

    for module in index['modules']:
        for ext in ['mli', 'ml']:
            name = f'{module.lower()}.{ext}'

            filename = os.path.join(path, name)
            if not os.path.exists(filename):
                continue

            with open(filename, 'r') as file:
                text = file.read()
                files.append((name, text))

    return files


def tar_to_files(path=None, raw=None) -> List[Tuple[str, str]]:
    assert path or raw

    fileobj = io.BytesIO(raw) if raw else None
    with tarfile.open(name=path, fileobj=fileobj) as tar:
        with TemporaryDirectory() as tmp_dir:
            tar.extractall(tmp_dir)
            files = dir_to_files(tmp_dir)

    return files


def url_to_files(url) -> List[Tuple[str, str]]:
    res = requests.get(url, stream=True)
    raw = b''

    for data in tqdm(res.iter_content()):
        raw += data

    return tar_to_files(raw=raw)


def files_to_proto(files: List[Tuple[str, str]]) -> dict:
    components = OrderedDict()

    for filename, text in files:
        name, ext = filename.split('.')
        key = {'mli': 'interface', 'ml': 'implementation'}[ext]
        name = name.capitalize()
        data = hexlify(text.encode()).decode()

        if name in components:
            components[name][key] = data
        else:
            components[name] = {'name': name, key: data}

    proto = {
        'expected_env_version': 0,  # TODO: this is V1
        'components': list(components.values())
    }
    return proto


def files_to_tar(files: List[Tuple[str, str]], output_path=None):
    fileobj = io.BytesIO() if output_path is None else None
    nameparts = os.path.basename(output_path).split('.')
    mode = 'w'
    if len(nameparts) == 3:
        mode = f'w:{nameparts[-1]}'

    with tarfile.open(name=output_path, fileobj=fileobj, mode=mode) as tar:
        for filename, text in files:
            file = io.BytesIO(text.encode())
            ti = tarfile.TarInfo(filename)
            ti.size = len(file.getvalue())
            tar.addfile(ti, file)

    if fileobj:
        return fileobj.getvalue()


def proto_to_files(proto: dict) -> List[Tuple[str, str]]:
    files = list()
    extensions = {'interface': 'mli', 'implementation': 'ml'}

    for component in proto.get('components', []):
        for key, ext in extensions.items():
            if key in component:
                filename = f'{component["name"].lower()}.{ext}'
                text = bytes.fromhex(component[key]).decode()
                files.append((filename, text))

    return files


def proto_to_bytes(proto: dict) -> bytes:
    res = b''

    for component in proto.get('components', []):
        res += netstruct.pack(b'I$', component['name'].encode())

        if component.get('interface'):
            res += b'\xff' + netstruct.pack(b'I$', bytes.fromhex(component['interface']))
        else:
            res += b'\x00'

        # we should also handle patch case
        res += netstruct.pack(b'I$', bytes.fromhex(component.get('implementation', '')))

    res = netstruct.pack(b'hI$', proto['expected_env_version'], res)
    return res


class Proto(metaclass=InlineDocstring):

    def __init__(self, proto):
        self._proto = proto

    def __repr__(self):
        res = [
            super(Proto, self).__repr__(),
            '\nHelpers',
            get_class_docstring(self.__class__)
        ]
        return '\n'.join(res)

    def __iter__(self):
        return iter(proto_to_files(self._proto))

    @classmethod
    def from_uri(cls, uri):
        """ Loads protocol implementation from various sources and converts it to the RPC-like format.

        :param uri: link/path to a tar archive or path to a folder with extracted contents
        :returns: Protocol instance
        """
        if uri.startswith('http'):
            files = url_to_files(uri)
        elif os.path.exists(os.path.expanduser(uri)):
            files = tar_to_files(uri)
        elif os.path.isdir(uri):
            files = dir_to_files(uri)
        else:
            raise ValueError(uri)

        return Proto(files_to_proto(files))

    def index(self) -> dict:
        """ Generates TEZOS_PROTOCOL file.

        :returns: dict with protocol hash and modules
        """
        data = {
            'hash': self.hash(),
            'modules': list(map(lambda x: x['name'], self._proto.get('components', [])))
        }
        return data

    def export_tar(self, output_path=None):
        """ Creates a tarball and dumps to a file or returns bytes.

        :param output_path: Path to the tarball [optional]. You can add .bz2 or .gz extension to make it compressed
        :returns: bytes if path is None or nothing
        """
        files = proto_to_files(self._proto)
        files.append(('TEZOS_PROTOCOL', json.dumps(self.index())))
        return files_to_tar(files, output_path)

    def export_html(self, output_path=None):
        """ Generates github-like side-by-side diff viewe, powered by diff2html.js

        :param output_path: will write to this file if specified
        :returns: html string if path is not specified
        """
        diffs = [text for filename, text in self if text]
        return generate_unidiff_html(diffs, output_path=output_path)

    def diff(self, proto, context_size=3):
        """ Calculates file diff between two protocol versions.

        :param proto: an instance of Protocol
        :param context_size: number of context lines before and after the change
        :returns: patch in proto format
        """
        files = list()
        yours = dict(iter(self))
        theirs = proto_to_files(proto())

        for filename, their_text in theirs:
            patch = make_patch(
                a=yours.get(filename, ''),
                b=their_text,
                filename=filename,
                context_size=context_size
            )
            files.append((filename, patch))

        return Proto(files_to_proto(files))

    def patch(self, patch):
        """ Applies unified diff and returns full-fledged protocol.

        :param patch: an instance of Protocol containing diff of files
        :returns: Protocol instance
        """
        files = list()
        yours = dict(iter(self))
        diff = proto_to_files(patch())

        for filename, diff_text in diff:
            text = yours.get(filename, '')
            if diff_text:
                text = apply_patch(text, diff_text)
            files.append((filename, text))

        return Proto(files_to_proto(files))

    def hash(self):
        hash_digest = blake2b_32(proto_to_bytes(self._proto)).digest()
        return base58_encode(hash_digest, b'P').decode()



from typing import Any, Callable, Generator
from loguru import logger

from pytezos.rpc.node import RpcError
from pytezos.rpc.query import RpcQuery
from pytezos.tools.docstring import get_attr_docstring
from pytezos.encoding import is_bh


def find_state_change_intervals(head: int, last: int, get: Callable, equals: Callable,
                                step=60) -> Generator:
    succ_value = get(head)
    logger.debug(f'{succ_value} at head {head}')

    for level in range(head - step, last, -step):
        value = get(level)
        logger.debug(f'{value} at level {level}')

        if not equals(value, succ_value):
            logger.debug(f'{value} -> {succ_value} at ({level}, {level + step})')
            yield level + step, succ_value, level, value
            succ_value = value


def find_state_change(head: int, last: int, get: Callable, equals: Callable,
                      pred_value: Any) -> (int, Any):
    def bisect(start: int, end: int):
        if end == start + 1:
            return end, get(end)

        level = (end + start) // 2
        value = get(level)
        logger.debug(f'{value} at level {level}')

        if equals(value, pred_value):
            return bisect(level, end)
        else:
            return bisect(start, level)

    return bisect(last, head)


def walk_state_change_interval(head: int, last: int, get: Callable, equals: Callable,
                               head_value: Any, last_value: Any) -> Generator:
    level = last
    value = last_value
    while not equals(value, head_value):
        level, value = find_state_change(head, level, get, equals, pred_value=value)
        logger.debug(f'{last_value} -> {value} at {level}')
        yield level, value


def find_state_changes(head: int, last: int, get: Callable, equals: Callable,
                       step=60) -> Generator:
    state_change_intervals = find_state_change_intervals(head, last, get, equals, step)
    for int_head, int_head_value, int_tail, int_last_value in state_change_intervals:
        for change in walk_state_change_interval(int_head, int_tail, get, equals,
                                                 head_value=int_head_value,
                                                 last_value=int_last_value):
            yield change


class BlockSliceQuery(RpcQuery):

    def __init__(self, start: int, stop=None, **kwargs):
        super(BlockSliceQuery, self).__init__(**kwargs)
        self._start = start
        self._stop = stop or 'head'

    def __repr__(self):
        res = [
            super(BlockSliceQuery, self).__repr__(),
            f'\nBlock range\n`{self._start}` — `{self._stop}`',
            f'\n(){get_attr_docstring(BlockSliceQuery, "__call__")}'
        ]
        return '\n'.join(res)

    def __getitem__(self, item):
        """ Get block by index.

        :param item: Index inside given block range
        :rtype: BlockQuery
        """
        start, stop = self.get_range()
        if item >= 0:
            return self._getitem(start + item)
        else:
            return self._getitem(stop + item + 1)

    def __call__(self) -> list:
        """ Get block hashes (base58) for this interval.
        """
        if is_bh(self._stop):
            head = self._stop
        else:
            head = self._getitem(self._stop).hash()

        if self._start < 0:
            length = abs(self._start)
        else:
            header = self._getitem(self._stop).header()
            length = header['level'] - self._start + 1

        return super(BlockSliceQuery, self).__call__(length=length, head=head)

    def get_range(self):
        """ Get block level range.
        """
        def get_level(x):
            if isinstance(x, int):
                if x < 0:
                    return self.head.level() + x
                elif x > 0:
                    return x
                else:
                    assert False
            else:
                return self._getitem(x).header()['level']

        return get_level(self._start), get_level(self._stop)

    def find_proposal_injection(self, proposal_id):
        """ Find proposal injection.

        :param proposal_id: Proposal hash (base58)
        """
        last, head = self.get_range()
        level, _ = find_state_change(
            head=head - 1,  # proposals are empty at the last block
            last=last,
            get=lambda x: self._getitem(x).votes.proposals[proposal_id](),
            equals=lambda x, y: x == y,
            pred_value=0
        )
        votes = self._getitem(level).operations.find_votes(proposal_id)
        assert len(votes) == 1
        return votes

    def find_upvotes(self, proposal_id) -> Generator:
        """ Find upvoting operations for the given proposal.

        :param proposal_id: Proposal hash (base58)
        :returns: Generator (lazy)
        """
        last, head = self.get_range()
        state_changes = find_state_changes(
            head=head - 1,  # proposals are empty at the last block
            last=last,
            get=lambda x: self._getitem(x).votes.proposals[proposal_id](),
            equals=lambda x, y: x == y
        )
        for level, _ in state_changes:
            for upvote in self._getitem(level).operations.find_upvotes(proposal_id):
                yield upvote

    def find_ballots(self) -> Generator:
        """ Find ballot operations for the current period.

        :returns: Generator (lazy)
        """
        last, head = self.get_range()
        state_changes = find_state_changes(
                head=head - 1,  # ballots are empty at the last block
                last=last,
                get=lambda x: self._getitem(x).votes.ballots(),
                equals=lambda x, y: x == y
        )
        for level, _ in state_changes:
            for ballot in self._getitem(level).operations.find_ballots():
                yield ballot

    def find_origination(self, contract_id):
        """ Find contract origination.

        :param contract_id: Contract ID (KT-address)
        """
        def get_counter(x):
            try:
                return self._getitem(x).context.contracts[contract_id].counter()
            except RpcError:
                return None

        level, _ = find_state_change(
            head=self.head.level(),
            last=0,
            get=get_counter,
            equals=lambda x, y: x == y,
            pred_value=None
        )
        return self._getitem(level).operations.find_origination(contract_id)

    def find_operation(self, operation_group_hash) -> dict:
        """ Find operation by hash.

        :param operation_group_hash: base58
        :raises: StopIteration if not found
        """
        last, head = self.get_range()
        for block_level in range(head, max(1, last - 1), -1):
            try:
                return self._getitem(block_level).operations[operation_group_hash]()
            except StopIteration:
                continue

        raise StopIteration(operation_group_hash)

    
class PeriodQuery(RpcQuery):
    __pos_key__ = ''
    __val_key__ = ''

    def _get_item(self, item) -> BlockSliceQuery:
        lvl = self.head.metadata()['level']
        blocks_per_period = int((lvl['level'] - lvl[self.__pos_key__] - 1) / lvl[self.__val_key__])

        def get_range(x):
            if x > 0:
                start_lvl = (x - 1) * blocks_per_period + 1
            elif x < 0:
                start_lvl = (lvl[self.__val_key__] + x + 1) * blocks_per_period + 1
            else:
                assert False

            stop_lvl = start_lvl + blocks_per_period - 1
            if stop_lvl > lvl['level']:
                stop_lvl = None

            return start_lvl, stop_lvl

        if isinstance(item, slice):
            start, _ = get_range(item.start or 1)
            _, stop = get_range(item.stop or -1)
        elif isinstance(item, int):
            start, stop = get_range(item)
        else:
            raise NotImplementedError(item)

        return BlockSliceQuery(
            start=start,
            stop=stop,
            node=self.node,
            path=self._wild_path,
            params=self._params
        )


class CyclesQuery(PeriodQuery):
    __pos_key__ = 'cycle_position'
    __val_key__ = 'cycle'

    def __call__(self, **params):
        """ Get current cycle.
        """
        return self.head.cycle()

    def __getitem__(self, item) -> BlockSliceQuery:
        """ Get block range by cycle/cycle range.

        :param item: Cycle number or range (slice), range start/stop can be empty or negative
        :rtype: BlockSliceQuery
        """
        return self._get_item(item)


class VotingPeriodsQuery(PeriodQuery):
    __pos_key__ = 'voting_period_position'
    __val_key__ = 'voting_period'

    def __call__(self, **params):
        """ Get current voting period.
        """
        return self.head.voting_period()

    def __getitem__(self, item) -> BlockSliceQuery:
        """ Get block range by voting_period/voting_period range.

        :param item: Voting_period number or range (slice), range start/stop can be empty or negative
        :rtype: BlockSliceQuery
        """
        return self._get_item(item)



from hashlib import sha256, sha512

from pytezos import crypto as crypto
from pytezos.repl.control import instruction
from pytezos.repl.context import Context
from pytezos.repl.types import assert_stack_type, Int, Nat, Timestamp, Mutez, Option, Pair, Bool, Bytes, Key, \
    Signature, KeyHash, dispatch_type_map


@instruction('ABS')
def do_abs(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Int)
    res = Nat(abs(int(top)))
    ctx.push(res, annots=annots)


@instruction('ADD')
def do_add(ctx: Context, prim, args, annots):
    a, b = ctx.pop2()
    res_type = dispatch_type_map(a, b, {
        (Nat, Nat): Nat,
        (Nat, Int): Int,
        (Int, Nat): Int,
        (Int, Int): Int,
        (Timestamp, Int): Timestamp,
        (Int, Timestamp): Timestamp,
        (Mutez, Mutez): Mutez
    })
    res = res_type(int(a) + int(b))
    ctx.push(res, annots=annots)


@instruction('COMPARE')
def do_compare(ctx: Context, prim, args, annots):
    a, b = ctx.pop2()
    res = Int(a.__cmp__(b))
    ctx.push(res, annots=annots)


@instruction('EDIV')
def do_ediv(ctx: Context, prim, args, annots):
    a, b = ctx.pop2()
    q_type, r_type = dispatch_type_map(a, b, {
        (Nat, Nat): (Nat, Nat),
        (Nat, Int): (Int, Nat),
        (Int, Nat): (Int, Nat),
        (Int, Int): (Int, Nat),
        (Mutez, Nat): (Mutez, Mutez),
        (Mutez, Mutez): (Nat, Mutez)
    })
    if int(b) == 0:
        res = Option.none(Pair.new(q_type(), r_type()).type_expr)
    else:
        q, r = divmod(int(a), int(b))
        if r < 0:
            r += abs(int(b))
            q += 1
        res = Option.some(Pair.new(q_type(q), r_type(r)))
    ctx.push(res, annots=annots)


@instruction(['EQ', 'GE', 'GT', 'LE', 'LT', 'NEQ'])
def do_eq(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Int)
    handlers = {
        'EQ': lambda x: x == 0,
        'GE': lambda x: x >= 0,
        'GT': lambda x: x > 0,
        'LE': lambda x: x <= 0,
        'LT': lambda x: x < 0,
        'NEQ': lambda x: x != 0
    }
    res = Bool(handlers[prim](int(top)))
    ctx.push(res, annots=annots)


@instruction('INT')
def do_int(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Nat)
    res = Int(int(top))
    ctx.push(res, annots=annots)


@instruction('ISNAT')
def do_is_nat(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Int)
    if int(top) >= 0:
        res = Option.some(Nat(int(top)))
    else:
        res = Option.none(Nat().type_expr)
    ctx.push(res, annots=annots)


@instruction(['LSL', 'LSR'])
def do_lsl(ctx: Context, prim, args, annots):
    a, b = ctx.pop2()
    assert_stack_type(a, Nat)
    assert_stack_type(b, Nat)
    assert int(b) < 257, f'shift overflow {int(b)}, should not exceed 256'
    handlers = {
        'LSL': lambda x: x[0] << x[1],
        'LSR': lambda x: x[0] >> x[1]
    }
    res = Nat(handlers[prim]((int(a), int(b))))
    ctx.push(res, annots=annots)


@instruction('MUL')
def do_mul(ctx: Context, prim, args, annots):
    a, b = ctx.pop2()
    res_type = dispatch_type_map(a, b, {
        (Nat, Nat): Nat,
        (Nat, Int): Int,
        (Int, Nat): Int,
        (Int, Int): Int,
        (Mutez, Nat): Mutez,
        (Nat, Mutez): Mutez
    })
    res = res_type(int(a) * int(b))
    ctx.push(res, annots=annots)


@instruction('NEG')
def do_neg(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, [Int, Nat])
    res = Int(-int(top))
    ctx.push(res, annots=annots)


@instruction('SUB')
def do_sub(ctx: Context, prim, args, annots):
    a, b = ctx.pop2()
    res_type = dispatch_type_map(a, b, {
        (Nat, Nat): Int,
        (Nat, Int): Int,
        (Int, Nat): Int,
        (Int, Int): Int,
        (Timestamp, Int): Timestamp,
        (Timestamp, Timestamp): Int,
        (Mutez, Mutez): Mutez
    })
    res = res_type(int(a) - int(b))
    ctx.push(res, annots=annots)


@instruction(['OR', 'XOR'])
def do_or_xor(ctx: Context, prim, args, annots):
    a, b = ctx.pop2()
    val_type, res_type = dispatch_type_map(a, b, {
        (Bool, Bool): (bool, Bool),
        (Nat, Nat): (int, Nat),
    })
    handlers = {
        'OR': lambda x: x[0] | x[1],
        'XOR': lambda x: x[0] ^ x[1]
    }
    res = res_type(handlers[prim]((val_type(a), val_type(b))))
    ctx.push(res, annots=annots)


@instruction('AND')
def do_and(ctx: Context, prim, args, annots):
    a, b = ctx.pop2()
    val_type, res_type = dispatch_type_map(a, b, {
        (Bool, Bool): (bool, Bool),
        (Nat, Nat): (int, Nat),
        (Int, Nat): (int, Nat),
        (Nat, Int): (int, Nat)
    })
    res = res_type(val_type(a) & val_type(b))
    ctx.push(res, annots=annots)


@instruction('NOT')
def do_not(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, [Nat, Int, Bool])
    if type(top) in [Nat, Int]:
        res = Int(~int(top))
    elif type(top) == Bool:
        res = Bool(not bool(top))
    else:
        assert False
    ctx.push(res, annots=annots)


@instruction('BLAKE2B')
def do_blake2b(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Bytes)
    res = Bytes(crypto.blake2b_32(bytes(top)).digest())
    ctx.push(res, annots=annots)


@instruction('CHECK_SIGNATURE')
def do_check_sig(ctx: Context, prim, args, annots):
    pk, sig, msg = ctx.pop3()
    assert_stack_type(pk, Key)
    assert_stack_type(sig, Signature)
    assert_stack_type(msg, Bytes)
    key = crypto.Key.from_encoded_key(str(pk))
    try:
        key.verify(signature=str(sig), message=bytes(msg))
    except:
        res = Bool(False)
    else:
        res = Bool(True)
    ctx.push(res, annots=annots)


@instruction('HASH_KEY')
def do_hash_key(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Key)
    key = crypto.Key.from_encoded_key(str(top))
    res = KeyHash.new(key.public_key_hash())
    ctx.push(res, annots=annots)


@instruction(['SHA256', 'SHA512'])
def do_sha(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Bytes)
    handlers = {
        'SHA256': lambda x: sha256(x).digest(),
        'SHA512': lambda x: sha512(x).digest(),
    }
    res = Bytes(handlers[prim](bytes(top)))
    ctx.push(res, annots=annots)



from loguru import logger

from pytezos.encoding import is_pkh
from pytezos.crypto import Key
from pytezos.rpc import mainnet


class OTP:

    def __init__(self, key, interval=5, shell=mainnet):
        """
        :param key: secret key (encrypted/unencrypted), public key or public key hash, all base58 encoded
        :param interval: number of blocks to check (tolerance)
        :param shell: ShellQuery instance
        """
        if not isinstance(key, Key):
            if is_pkh(key):
                key = shell.public_key(key)
            key = Key.from_encoded_key(key)

        self._key = key
        self._interval = interval
        self._shell = shell

    def now(self) -> str:
        if not self._key.secret_exponent:
            raise ValueError('Cannot generate OTP without a secret key')

        message = self._shell.head.calculate_hash()
        logger.debug(f'block hash: {message}')

        return self._key.sign(message)

    def verify(self, signature) -> bool:
        block_hashes = self._shell.blocks(length=self._interval)

        for row in block_hashes:
            try:
                message = row[0]
                logger.debug(f'try {message}')

                self._key.verify(signature, message)
            except ValueError as e:
                logger.debug(str(e))
            else:
                return True

        return False



from pprint import pprint
from collections import namedtuple

from pytezos.michelson.forge import prim_tags
from pytezos.michelson.micheline import Schema, collapse_micheline, build_maps, parse_micheline, \
    parse_json, make_micheline, michelson_to_micheline
from pytezos.michelson.formatter import micheline_to_michelson
from pytezos.michelson.docstring import generate_docstring

BigMapSchema = namedtuple('BigMapSchema', ['bin_to_id', 'id_to_bin'])


class MichelineSchemaError(ValueError):
    pass


def build_schema(code) -> Schema:
    """ Creates internal structures necessary for decoding/encoding micheline:
    `metadata` -> micheline tree with collapsed `pair`, `or`, and `option` nodes
    `bin_types` -> maps binary path to primitive
    `bin_names` -> binary path to key name mapping
    `json_to_bin` -> json path to binary path mapping

    :param code: parameter or storage section of smart contract source code (in micheline)
    :rtype: Schema
    """
    try:
        metadata = collapse_micheline(code)
        return Schema(metadata, *build_maps(metadata))
    except (KeyError, ValueError, TypeError) as e:
        pprint(code, compact=True)
        raise MichelineSchemaError(f'Failed to build schema', e.args)


def decode_micheline(val_expr, type_expr, schema: Schema, root='0'):
    """ Converts Micheline data into Python object.

    :param val_expr: Micheline value expression
    :param type_expr: Michelson type expression for the entire type
    :param schema: schema built for particular contract/section
    :param root: which binary node to take as root, used to decode BigMap values/diffs
    :returns: Object
    """
    try:
        return parse_micheline(val_expr, type_expr, schema, root)
    except (KeyError, IndexError, TypeError) as e:
        print(generate_docstring(schema, 'schema'))
        pprint(val_expr, compact=True)
        raise MichelineSchemaError(f'Failed to decode micheline expression', e.args)


def encode_micheline(data, schema: Schema, root='0', binary=False):
    """ Converts Python object into Micheline expression.

    :param data: Python object
    :param schema: schema built for particular contract/section
    :param root: which binary node to take as root, used to encode BigMap values
    :param binary: Encode keys and addresses in bytes rather than strings, default is False
    :returns: Micheline expression
    """
    try:
        bin_values = parse_json(data, schema, root)
        return make_micheline(bin_values, schema.bin_types, root, binary)
    except (KeyError, IndexError, TypeError) as e:
        print(generate_docstring(schema, 'schema'))
        pprint(data, compact=True)
        raise MichelineSchemaError(f'Failed to encode micheline expression', e.args)


def convert(source, schema: Schema = None, output='micheline', inline=False):
    """ Convert data between different representations (DO NOT USE FOR STORAGE/PARAMETER, can be ambiguous).

    :param source: Data, can be one of Michelson (string), Micheline expression, object
    :param schema: Needed if decoding/encoding objects (optional)
    :param output: Output format, one of 'micheline' (default), 'michelson', 'object'
    :param inline: Used for michelson output, whether to omit line breaks
    """
    if isinstance(source, str):
        try:
            source = michelson_to_micheline(source)
        except ValueError:
            assert schema
            source = encode_micheline(source, schema)
    elif not is_micheline(source):
        assert schema
        source = encode_micheline(source, schema)

    if output == 'michelson':
        return micheline_to_michelson(source, inline)
    elif output == 'object':
        assert False, f'not supported'
    elif output == 'micheline':
        return source
    else:
        assert False, output


def build_big_map_schema(data, schema: Schema) -> BigMapSchema:
    """ Generate Big_map schema from the contract storage

    :param data: Raw storage (Micheline expression)
    :param schema: Storage schema
    :returns: Mappings: Big_map id to JSON path and vice versa
    :rtype: BigMapSchema
    """
    bin_to_id = dict()
    id_to_bin = dict()

    def scan_big_map_ids(node, path):
        if len(path) == 0:
            assert node.get('int'), (node, path)
            yield int(node['int'])
        elif isinstance(node, list):
            for item in node:
                yield from scan_big_map_ids(item, path)
        else:
            assert node.get('args'), (node, path)
            yield from scan_big_map_ids(node['args'][int(path[0])], path[1:])

    for bin_path, prim in schema.bin_types.items():
        if prim == 'big_map':
            for big_map_id in scan_big_map_ids(data, bin_path[1:]):
                bin_to_id[bin_path], id_to_bin[big_map_id] = big_map_id, bin_path

    return BigMapSchema(bin_to_id, id_to_bin)


def is_micheline(value) -> bool:
    """ Check if value is a Micheline expression (using heuristics, so not 100% accurate).

    :param value: Object
    :rtype: bool
    """
    if isinstance(value, list):
        def get_prim(x):
            return x.get('prim') if isinstance(x, dict) else None
        return set(map(get_prim, value)) == {'parameter', 'storage', 'code'}
    elif isinstance(value, dict):
        primitives = list(prim_tags.keys())
        return any(map(lambda x: x in value, ['prim', 'args', 'annots', *primitives]))
    else:
        return False



import re
import inspect
import types
from functools import update_wrapper


def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')


__interactive_mode__ = is_interactive()


def     get_attr_docstring(class_type, attr_name):
    if attr_name == 'get':
        attr_name = '__call__'

    attr = getattr(class_type, attr_name, None)
    if attr and attr.__doc__:
        return re.sub(r' {3,}', '', attr.__doc__)


def default_attr_filter(x):
    return not x.startswith('_')


def get_class_docstring(class_type, attr_filter=default_attr_filter, extended=False):
    def attr_format(x):
        attr = getattr(class_type, x)

        if type(attr) == property:
            name = f'.{x}'
        else:
            if extended:
                sig = str(inspect.signature(attr)).replace('self, ', '')
            else:
                sig = '()'
            name = f'.{x}{sig}'

        if extended:
            doc = get_attr_docstring(class_type, x)
        else:
            doc = ''

        return f'{name}{doc}'

    return '\n'.join(map(attr_format, filter(attr_filter, dir(class_type))))


def inline_doc(method):
    if not __interactive_mode__:
        return method

    doc = [repr(method)]
    if method.__doc__:
        doc.append(re.sub(r' {3,}', '', method.__doc__))

    class CustomReprDescriptor:
        def __get__(self, instance, owner):
            class MethodWrapper:
                def __init__(self):
                    self.class_instance = instance
                    self.doc = '\n'.join(doc)

                def __call__(self, *args, **kwargs):
                    return method(self.class_instance, *args, **kwargs)

                def __repr__(self):
                    return self.doc

            return update_wrapper(MethodWrapper(), method)

    return CustomReprDescriptor()


class InlineDocstring(type):

    def __new__(mcs, name, bases, attrs, **kwargs):
        new_attrs = {}
        for attr_name, attr in attrs.items():
            if isinstance(attr, types.FunctionType) and attr.__doc__ and not attr_name.startswith('_'):
                attr = inline_doc(attr)
            new_attrs[attr_name] = attr
        return type.__new__(mcs, name, bases, new_attrs, **kwargs)



import functools
from typing import Tuple, Callable

from pytezos.michelson.formatter import micheline_to_michelson
import pytezos.encoding as encoding

parsers = {}


def primitive(prim, args_len=0):
    def register_primitive(func):
        parsers[prim] = (args_len, func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return register_primitive


class Unit(object):

    def __repr__(self):
        return 'Unit'

    def __eq__(self, other):
        return isinstance(other, Unit)


class MichelsonRuntimeError(ValueError):

    def __init__(self, message, trace):
        super(MichelsonRuntimeError, self).__init__(f'{message}: {" -> ".join(trace)}')
        self.message = message
        self.trace = trace

    @classmethod
    def init(cls, message, prim):
        return cls(message, trace=[prim])

    @classmethod
    def wrap(cls, error: 'MichelsonRuntimeError', prim):
        return cls(error.message, trace=[prim] + error.trace)


class MichelsonTypeCheckError(MichelsonRuntimeError):
    pass


def assert_type(value, exp_type):
    assert isinstance(value, exp_type), f'expected {exp_type.__name__}, got {type(value).__name__}'


def parse_prim_expr(expr) -> tuple:
    assert isinstance(expr, dict), f'expected dict, got {type(expr).__name__}'
    assert 'prim' in expr, f'prim field is absent'
    return expr['prim'], expr.get('args', [])


def get_prim_args(val_expr, prim, args_len: int):
    p, args = parse_prim_expr(val_expr)
    assert p == prim, f'expected {prim}, got {p}'
    assert len(args) == args_len, f'expected {args_len} args, got {len(args)}'
    return args


def dispatch_prim_map(val_expr, mapping: dict):
    p, args = parse_prim_expr(val_expr)
    expected = ' or '.join(map(lambda x: f'{x[0]} ({x[1]} args)', mapping.items()))
    assert (p, len(args)) in mapping, f'expected {expected}, got {p} ({len(args)} args)'
    res = mapping[(p, len(args))]
    if callable(res):
        return res(args)
    else:
        return res


def parse_core_expr(expr) -> tuple:
    assert isinstance(expr, dict), f'expected dict, got {type(expr)}'
    core_type, value = next((k, v) for k, v in expr.items() if k[0] != '_' and k != 'annots')
    return core_type, expr[core_type]


def dispatch_core_map(val_expr, mapping: dict):
    act_type, val = parse_core_expr(val_expr)
    expected = ' or '.join(map(lambda x: f'`{x}`', mapping))
    assert act_type in mapping, f'expected {expected}, got {act_type}'
    res = mapping[act_type]
    if callable(res):
        try:
            return res(val)
        except ValueError as e:
            raise MichelsonTypeCheckError.init(str(e), act_type)
    else:
        return res


def get_core_val(val_expr, core_type):
    act_type, value = parse_core_expr(val_expr)
    assert core_type == act_type, f'expected {core_type}, got {act_type}'
    return value


def get_string(val_expr):
    return get_core_val(val_expr, core_type='string')


def get_int(val_expr):
    return int(get_core_val(val_expr, core_type='int'))


def get_bool(val_expr):
    return dispatch_prim_map(val_expr, {('True', 0): True, ('False', 0): False})


def get_bytes(val_expr):
    return bytes.fromhex(get_core_val(val_expr, core_type='bytes'))


def get_entry_expr(type_expr, field_annot):
    def _get(node, path):
        assert_type(node, dict)
        if field_annot in node.get('annots', []):
            return node, path

        for i, arg in enumerate(node.get('args', [])):
            res = _get(arg, path + str(i))
            if res:
                return res

    entry = _get(type_expr, '')
    if not entry and field_annot == '%default':
        entry_type, entry_path = type_expr, ''
    else:
        entry_type, entry_path = entry

    assert entry_type, f'entrypoint `{field_annot[1:]}` was not found'
    return entry_type, entry_path


def restore_entry_expr(val_expr, type_expr, field_annot):
    _, entry_path = get_entry_expr(type_expr, field_annot)
    for idx in reversed(entry_path):
        val_expr = {'prim': 'Left' if idx == '0' else 'Right',
                    'args': [val_expr]}
    return val_expr


def expr_equal(a, b):
    if type(a) != type(b):
        return False
    elif isinstance(a, dict):
        if a.get('prim'):
            if a['prim'] != b['prim']:
                return False
            else:
                return expr_equal(a.get('args', []), b.get('args', []))
        else:
            return a == b
    elif isinstance(a, list):
        if len(a) != len(b):
            return False
        else:
            return all(map(lambda i: expr_equal(a[i], b[i]), range(len(a))))
    else:
        assert False, (a, b)


def assert_expr_equal(expected, actual):
    assert expr_equal(expected, actual), \
        f'expected {micheline_to_michelson(expected)}, got {micheline_to_michelson(actual)}'


def is_comparable(type_expr):
    prim, args, _ = parse_type(type_expr)
    comparable_types = {
        'string', 'int', 'bytes', 'nat', 'bool',
        'address', 'key_hash', 'mutez', 'timestamp'
    }
    if prim in comparable_types:
        return True
    elif prim == 'pair':
        left, _ = parse_prim_expr(args[0])
        return left != 'pair' and all(map(is_comparable, args))
    else:
        return False


def assert_comparable(type_expr):
    assert is_comparable(type_expr), f'type is not comparable: {micheline_to_michelson(type_expr)}'


def is_pushable(type_expr):
    prim, args, _ = parse_type(type_expr)
    if prim in ['big_map', 'contract', 'operation']:
        return False
    elif prim == 'lambda':
        return True
    elif args:
        return all(map(is_pushable, args))
    else:
        return True


def assert_pushable(type_expr):
    assert is_pushable(type_expr), f'type is not pushable: {micheline_to_michelson(type_expr)}'


def is_big_map_val(type_expr):
    prim, args, _ = parse_type(type_expr)
    if prim in ['big_map', 'operation']:
        return False
    return all(map(is_big_map_val, args))


def assert_big_map_val(type_expr):
    assert is_big_map_val(type_expr), f'cannot be a big map value: {micheline_to_michelson(type_expr)}'


def parse_type(type_expr) -> Tuple[str, list, Callable]:
    prim, args = parse_prim_expr(type_expr)
    if prim not in parsers:
        raise MichelsonRuntimeError.init('unknown primitive', prim)
    args_len, func = parsers[prim]
    if len(args) != args_len:
        raise MichelsonRuntimeError.init(f'expected {args_len} arg(s), got {len(args)}', prim)
    return prim, args, func


def val_selector(val_expr, type_expr, val, type_path):
    prim = type_expr['prim']
    if prim == 'pair':
        return tuple(val)
    elif prim == 'option':
        return tuple(val) if val is not None else None
    elif prim == 'or':
        return val[0]
    elif prim == 'set':
        return set(val)
    elif prim == 'map':
        return dict(val)
    elif prim == 'big_map' and isinstance(val_expr, list):
        return dict(val)
    else:
        return val


def parse_expression(val_expr, type_expr, selector=val_selector, type_path='0'):
    """ Run an extensible parser for Micheline expressions.
    This function will just do the type checking for you,
    if you want to build a response you'd need to define a selector.

    :param val_expr: value expression (Micheline)
    :param type_expr: corresponding type expression (Micheline)
    :param selector: function selector(val_expr, type_expr, val, type_path). \
    Default selector (val_selector) converts Micheline to Python objects
    :param type_path: starting binary path (default is '0')
    """
    prim, _, func = parse_type(type_expr)
    try:
        res = func(val_expr, type_expr, selector, type_path)
    except AssertionError as e:
        raise MichelsonTypeCheckError.init(str(e), prim)
    except MichelsonRuntimeError as e:
        raise MichelsonTypeCheckError.wrap(e, prim)
    else:
        return res


@primitive('string')
def parse_string(val_expr, type_expr, selector, type_path):
    return selector(val_expr, type_expr, get_string(val_expr), type_path)


@primitive('int')
def parse_int(val_expr, type_expr, selector, type_path):
    return selector(val_expr, type_expr, get_int(val_expr), type_path)


@primitive('bytes')
def parse_bytes(val_expr, type_expr, selector, type_path):
    return selector(val_expr, type_expr, get_bytes(val_expr), type_path)


@primitive('nat')
def parse_nat(val_expr, type_expr, selector, type_path):
    val = get_int(val_expr)
    assert val >= 0, f'nat cannot be negative ({val})'
    return selector(val_expr, type_expr, val, type_path)


@primitive('bool')
def parse_bool(val_expr, type_expr, selector, type_path):
    val = dispatch_prim_map(val_expr, {
        ('False', 0): False,
        ('True', 0): True
    })
    return selector(val_expr, type_expr, val, type_path)


@primitive('unit')
def parse_unit(val_expr, type_expr, selector, type_path):
    _ = get_prim_args(val_expr, prim='Unit', args_len=0)
    return selector(val_expr, type_expr, Unit(), type_path)


@primitive('list', args_len=1)
def parse_list(val_expr, type_expr, selector, type_path):
    assert_type(val_expr, list)
    val = [parse_expression(item, type_expr['args'][0], selector, type_path + '0') for item in val_expr]
    return selector(val_expr, type_expr, val, type_path)


@primitive('pair', args_len=2)
def parse_pair(val_expr, type_expr, selector, type_path):
    args = get_prim_args(val_expr, prim='Pair', args_len=2)
    val = [parse_expression(args[i], type_expr['args'][i], selector, type_path + str(i)) for i in [0, 1]]
    return selector(val_expr, type_expr, val, type_path)


@primitive('option', args_len=1)
def parse_option(val_expr, type_expr, selector, type_path):
    val = dispatch_prim_map(val_expr, {
        ('None', 0): None,
        ('Some', 1): lambda x: [parse_expression(x[0], type_expr['args'][0], selector, type_path + '0')]
    })
    return selector(val_expr, type_expr, val, type_path)


@primitive('or', args_len=2)
def parse_or(val_expr, type_expr, selector, type_path):
    val = dispatch_prim_map(val_expr, {
        ('Left', 1): lambda x: [parse_expression(x[0], type_expr['args'][0], selector, type_path + '0')],
        ('Right', 1): lambda x: [parse_expression(x[0], type_expr['args'][1], selector, type_path + '1')]
    })
    return selector(val_expr, type_expr, val, type_path)


@primitive('set', args_len=1)
def parse_set(val_expr, type_expr, selector, type_path):
    assert_type(val_expr, list)
    assert_comparable(type_expr['args'][0])
    val = [parse_expression(item, type_expr['args'][0], selector, type_path + '0') for item in val_expr]
    return selector(val_expr, type_expr, val, type_path)


def parse_elt(val_expr, type_expr, selector, type_path):
    elt_args = get_prim_args(val_expr, prim='Elt', args_len=2)
    return [parse_expression(elt_args[i], type_expr['args'][i], selector, type_path + str(i)) for i in [0, 1]]


@primitive('map', args_len=2)
def parse_map(val_expr, type_expr, selector, type_path):
    assert_type(val_expr, list)
    assert_comparable(type_expr['args'][0])
    val = [parse_elt(item, type_expr, selector, type_path) for item in val_expr]
    return selector(val_expr, type_expr, val, type_path)


@primitive('big_map', args_len=2)
def parse_big_map(val_expr, type_expr, selector, type_path):
    assert_comparable(type_expr['args'][0])
    assert_big_map_val(type_expr['args'][1])
    if isinstance(val_expr, list):
        return parse_map(val_expr, type_expr, selector, type_path)
    else:
        val = get_int(val_expr)
        return selector(val_expr, type_expr, val, type_path)


@primitive('timestamp')
def parse_timestamp(val_expr, type_expr, selector, type_path):
    val = dispatch_core_map(val_expr, {'int': int, 'string': encoding.forge_timestamp})
    return selector(val_expr, type_expr, val, type_path)


@primitive('mutez')
def parse_mutez(val_expr, type_expr, selector, type_path):
    return parse_nat(val_expr, type_expr, selector, type_path)


@primitive('address')
def parse_address(val_expr, type_expr, selector, type_path):
    val = dispatch_core_map(val_expr, {
        'bytes': lambda x: encoding.parse_address(bytes.fromhex(x)),
        'string': lambda x: x
    })
    return selector(val_expr, type_expr, val, type_path)


@primitive('contract', args_len=1)
def parse_contract(val_expr, type_expr, selector, type_path):
    val = dispatch_core_map(val_expr, {
        'bytes': lambda x: encoding.parse_contract(bytes.fromhex(x)),
        'string': lambda x: x
    })
    return selector(val_expr, type_expr, val, type_path)


@primitive('operation')
def parse_operation(val_expr, type_expr, selector, type_path):
    return selector(val_expr, type_expr, get_string(val_expr), type_path)


@primitive('key')
def parse_key(val_expr, type_expr, selector, type_path):
    val = dispatch_core_map(val_expr, {
        'bytes': lambda x: encoding.parse_public_key(bytes.fromhex(x)),
        'string': lambda x: x
    })
    return selector(val_expr, type_expr, val, type_path)


@primitive('key_hash')
def parse_key_hash(val_expr, type_expr, selector, type_path):
    return parse_address(val_expr, type_expr, selector, type_path)


@primitive('signature')
def parse_signature(val_expr, type_expr, selector, type_path):
    val = dispatch_core_map(val_expr, {
        'bytes': lambda x: encoding.parse_signature(bytes.fromhex(x)),
        'string': lambda x: x
    })
    return selector(val_expr, type_expr, val, type_path)


@primitive('chain_id')
def parse_chain_id(val_expr, type_expr, selector, type_path):
    val = dispatch_core_map(val_expr, {
        'bytes': lambda x: encoding.parse_chain_id(bytes.fromhex(x)),
        'string': lambda x: x
    })
    return selector(val_expr, type_expr, val, type_path)


@primitive('lambda', args_len=2)
def parse_lambda(val_expr, type_expr, selector, type_path):
    assert_type(val_expr, list)
    return selector(val_expr, type_expr, val_expr, type_path)



import functools
from pprint import pformat

from pytezos.repl.context import Context
from pytezos.repl.types import StackItem, assert_stack_type, assert_expr_equal, Option, Lambda, Bool, List, Or, Pair, \
    Map, Set
from pytezos.repl.parser import get_int, MichelsonRuntimeError, assert_pushable, parse_prim_expr

instructions = {}


def assert_no_annots(prim, annots):
    assert not annots, f'unexpected annotations {annots}'


def instruction(prim, args_len=0):
    def register_instruction(func):
        if isinstance(prim, list):
            for p in prim:
                instructions[(p, args_len)] = func
        else:
            instructions[(prim, args_len)] = func

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return register_instruction


def parse_instruction(code_expr):
    prim, args = parse_prim_expr(code_expr)
    key = (prim, len(args))
    if key not in instructions:
        raise MichelsonRuntimeError.init(f'unknown instruction or wrong args len: {key}', prim)
    handler = instructions[key]
    annots = code_expr.get('annots', [])
    return prim, args, annots, handler
    

def do_interpret(ctx: Context, code_expr):
    res = None
    if isinstance(code_expr, list):
        for item in code_expr:
            res = do_interpret(ctx, item)
    elif isinstance(code_expr, dict):
        prim, args, annots, handler = parse_instruction(code_expr)
        try:
            ctx.begin(prim)
            res = handler(ctx, prim, args, annots)
        except AssertionError as e:
            raise MichelsonRuntimeError.init(str(e), prim)
        except MichelsonRuntimeError as e:
            raise MichelsonRuntimeError.wrap(e, prim)
        finally:
            ctx.end()
    else:
        assert False, f'unexpected code expression {pformat(code_expr, compact=True)}'
    return res


@instruction('PUSH', args_len=2)
def do_push(ctx: Context, prim, args, annots):
    item = StackItem.parse(val_expr=args[1], type_expr=args[0])
    assert_pushable(item.type_expr)
    ctx.push(item, annots=annots)


@instruction('DROP', args_len=1)
def do_drop(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    count = get_int(args[0])
    _ = ctx.pop(count=count)


@instruction('DROP')
def do_drop_1(ctx: Context, prim, args, annots):
    do_drop(ctx, prim, [{'int': '1'}], annots)


@instruction('DUP')
def do_dup(ctx: Context, prim, args, annots):
    top = ctx.peek()
    ctx.push(top, annots=annots)


@instruction('SWAP')
def do_swap(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    a, b = ctx.pop2()
    ctx.push(a, move=True)
    ctx.push(b, move=True)


@instruction('DIG', args_len=1)
def do_dig(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    index = get_int(args[0])
    ctx.protect(index)
    res = ctx.pop1()
    ctx.restore(index)
    ctx.push(res, move=True)


@instruction('DUG', args_len=1)
def do_dug(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    index = get_int(args[0])
    res = ctx.pop1()
    ctx.protect(index)
    ctx.push(res, move=True)
    ctx.restore(index)


@instruction('DIP', args_len=2)
def do_dip(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    count = get_int(args[0])
    ctx.protect(count)
    do_interpret(ctx, args[1])
    ctx.restore(count)


@instruction('DIP', args_len=1)
def do_dip_1(ctx: Context, prim, args, annots):
    do_dip(ctx, prim, [{'int': '1'}, args[0]], annots)


@instruction('LAMBDA', args_len=3)
def do_lambda(ctx: Context, prim, args, annots):
    res = Lambda.new(p_type_expr=args[0], r_type_expr=args[1], code=args[2])
    ctx.push(res, annots=annots)


@instruction('EXEC')
def do_exec(ctx: Context, prim, args, annots):
    param, lmbda = ctx.pop2()
    assert_stack_type(lmbda, Lambda)
    lmbda.assert_param_type(param)
    lmbda_ctx = ctx.spawn(stack=[param])
    do_interpret(lmbda_ctx, lmbda.code)
    ret = lmbda_ctx.pop1()
    lmbda.assert_ret_type(ret)
    assert len(lmbda_ctx) == 0, f'lambda stack is not empty {lmbda_ctx}'
    ctx.push(ret, annots=annots)


@instruction('APPLY')
def do_apply(ctx: Context, prim, args, annots):
    param, lmbda = ctx.pop2()
    assert_stack_type(lmbda, Lambda)
    res = lmbda.partial_apply(param)
    ctx.push(res)


@instruction('FAILWITH')
def do_failwith(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    top = ctx.pop1()
    assert False, repr(top)


@instruction('IF', args_len=2)
def do_if(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    cond = ctx.pop1()
    assert_stack_type(cond, Bool)
    do_interpret(ctx, args[0 if bool(cond) else 1])


@instruction('IF_CONS', args_len=2)
def do_if_cons(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    top = ctx.pop1()
    assert_stack_type(top, List)
    if len(top) > 0:
        head, tail = top.cut_head()
        ctx.push(tail)
        ctx.push(head)
        do_interpret(ctx, args[0])
    else:
        do_interpret(ctx, args[1])


@instruction('IF_LEFT', args_len=2)
def do_if_left(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    top = ctx.pop1()  # type: Or
    assert_stack_type(top, Or)
    ctx.push(top.get_some())
    do_interpret(ctx, args[0 if top.is_left() else 1])


@instruction('IF_NONE', args_len=2)
def do_if_left(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    top = ctx.pop1()
    assert_stack_type(top, Option)
    if top.is_none():
        do_interpret(ctx, args[0])
    else:
        ctx.push(top.get_some())
        do_interpret(ctx, args[1])


@instruction('LOOP', args_len=1)
def do_loop(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    while True:
        top = ctx.pop1()
        assert_stack_type(top, Bool)
        if bool(top):
            do_interpret(ctx, args[0])
        else:
            break


@instruction('LOOP_LEFT', args_len=1)
def do_loop_left(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    while True:
        top = ctx.pop1()
        assert_stack_type(top, Or)
        ctx.push(top.get_some())
        if top.is_left():
            do_interpret(ctx, args[0])
        else:
            break


@instruction('MAP', args_len=1)
def do_map(ctx: Context, prim, args, annots):
    container = ctx.pop1()
    assert_stack_type(container, [List, Map])

    if type(container) == List:
        items = list()
        for item in container:
            ctx.push(item)
            do_interpret(ctx, args[0])
            ret = ctx.pop1()
            items.append(ret)
    elif type(container) == Map:
        items = list()
        for key, val in container:
            ctx.push(Pair.new(key, val))
            do_interpret(ctx, args[0])
            ret = ctx.pop1()
            items.append((key, ret))
    else:
        assert False

    if len(items) == 0:
        res = container
    else:
        res = type(container).new(items)
    ctx.push(res, annots=annots)


@instruction('ITER', args_len=1)
def do_map(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    container = ctx.pop1()
    inferred_annots = [f'@{container.name}.elt'] if container.name else None
    if type(container) in [List, Set]:
        for item in container:
            ctx.push(item, annots=inferred_annots)
            do_interpret(ctx, args[0])
    elif type(container) == Map:
        for key, val in container:
            ctx.push(Pair.new(key, val), annots=inferred_annots)
            do_interpret(ctx, args[0])
    else:
        assert False, f'unexpected type {type(container)}'


@instruction('CAST', args_len=1)
def do_cast(ctx: Context, prim, args, annots):
    assert_no_annots(prim, annots)
    top = ctx.pop1()
    assert_expr_equal(args[0], top.type_expr)
    top.type_expr = args[0]
    ctx.push(top)


@instruction('RENAME')
def do_rename(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    ctx.push(top, annots=annots)



from pytezos.encoding import forge_address, forge_bool, forge_nat, forge_array, forge_public_key, \
    forge_base58
from pytezos.michelson.forge import forge_entrypoint, forge_micheline, forge_script

operation_tags = {
    'endorsement': 0,
    'proposal': 5,
    'ballot': 6,
    'seed_nonce_revelation': 1,
    'double_endorsement_evidence': 2,
    'double_baking_evidence': 3,
    'activate_account': 4,
    'reveal': 107,
    'transaction': 108,
    'origination': 109,
    'delegation': 110
}


def forge_operation(content) -> bytes:
    """ Forge operation content (locally).

    :param content: {.., "kind": "transaction", ...}
    """
    encode_content = {
        'activate_account': forge_activate_account,
        'reveal': forge_reveal,
        'transaction': forge_transaction,
        'origination': forge_origination,
        'delegation': forge_delegation
    }
    encode_proc = encode_content.get(content['kind'])
    if not encode_proc:
        raise NotImplementedError(content['kind'])

    return encode_proc(content)


def forge_operation_group(operation_group):
    """ Forge operation group (locally).

    :param operation_group: {"branch": "B...", "contents": [], ...}
    """
    res = forge_base58(operation_group['branch'])
    res += b''.join(map(forge_operation, operation_group['contents']))
    return res


def forge_activate_account(content: dict):
    res = forge_nat(operation_tags[content['kind']])
    res += forge_base58(content['pkh'])
    res += bytes.fromhex(content['secret'])
    return res


def forge_reveal(content):
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))
    res += forge_public_key(content['public_key'])
    return res


def forge_transaction(content):
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))
    res += forge_nat(int(content['amount']))
    res += forge_address(content['destination'])

    if content.get('parameters'):
        res += forge_bool(True)
        res += forge_entrypoint(content['parameters']['entrypoint'])
        res += forge_array(forge_micheline(content['parameters']['value']))
    else:
        res += forge_bool(False)

    return res


def forge_origination(content):
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))
    res += forge_nat(int(content['balance']))

    if content.get('delegate'):
        res += forge_bool(True)
        res += forge_address(content['delegate'], tz_only=True)
    else:
        res += forge_bool(False)

    res += forge_script(content['script'])

    return res


def forge_delegation(content):
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))

    if content.get('delegate'):
        res += forge_bool(True)
        res += forge_address(content['delegate'], tz_only=True)
    else:
        res += forge_bool(False)

    return res



import functools
import operator
from typing import List
from pytezos.rpc.errors import RpcError


class OperationResult:
    """ Operation result representation + useful parsing helpers for operation group
    """

    def __init__(self, **props):
        self.props = props
        for key, value in props.items():
            setattr(self, key, value)

    def __repr__(self):
        res = [
            super(OperationResult, self).__repr__(),
            '\nProperties',
            *list(map(lambda x: f'.{x}', self.props))
        ]
        return '\n'.join(res)

    @staticmethod
    def iter_contents(operation_group: dict):
        """ Lazily iterate operation group contents including internal operations.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        :returns: generator
        """
        contents = operation_group.get('contents', [operation_group])
        for content in contents:
            yield {'internal': False, **content}
            internal_operation_results = content.get('metadata', {}).get('internal_operation_results', [])
            for result in internal_operation_results:
                yield {'internal': True, **result}

    @staticmethod
    def iter_results(operation_group: dict):
        """ Lazily iterate operation results including internal operation results.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        :returns: generator
        """
        for content in OperationResult.iter_contents(operation_group):
            if content['internal'] and content.get('result'):
                yield content['result']
            elif not content['internal'] and content.get('metadata', {}).get('operation_result'):
                yield content['metadata']['operation_result']

    @staticmethod
    def consumed_gas(operation_group) -> int:
        """ Get total consumed gas for an operation group (recursively).

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        """
        return sum(map(lambda x: int(x.get('consumed_gas', '0')),
                       OperationResult.iter_results(operation_group)))

    @staticmethod
    def paid_storage_size_diff(operation_group) -> int:
        """ Get total paid storage size diff for an operation group (recursively).

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        """
        return sum(map(lambda x: int(x.get('paid_storage_size_diff', '0')),
                       OperationResult.iter_results(operation_group)))

    @staticmethod
    def burned(operation_group) -> int:
        """ Get total burned (due to account allocations) for an operation group (recursively).

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        """
        return sum(map(
            lambda x: 257 if x.get('allocated_destination_contract') or x.get('originated_contracts') else 0,
            OperationResult.iter_results(operation_group)))

    @staticmethod
    def is_applied(operation_group) -> bool:
        """ Check if ALL operations in a group are applied.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        """
        return all(map(lambda x: x['status'] == 'applied',
                       OperationResult.iter_results(operation_group)))

    @staticmethod
    def errors(operation_group: dict) -> list:
        """ Collect errors from all operation results in a group.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        :returns: list of errors [{"id": "", ...}]
        """
        all_errors = (
            result.get("errors", []) if result["status"] != "applied" else []
            for result in OperationResult.iter_results(operation_group)
        )
        return functools.reduce(operator.iconcat, all_errors, [])

    @staticmethod
    def originated_contracts(operation_group: dict) -> list:
        """ Collect originated contract addresses from all operation results in a group.

        :param operation_group: {"branch": "B...", "contents": [...], ...} \
        OR a single content {"kind": "transaction", ...}
        :returns: list of addresses ["tz12345...", ...]
        """
        originated_contracts = list()
        for result in OperationResult.iter_results(operation_group):
            originated_contracts.extend(result.get('originated_contracts', []))
        return originated_contracts

    @staticmethod
    def get_contents(operation_group: dict, **predicates):
        def match(x):
            return all(map(lambda pred: x.get(pred[0]) == pred[1], predicates.items()))

        if not predicates:
            assert len(operation_group['contents']) == 1
            return operation_group['contents']
        else:
            return list(filter(match, OperationResult.iter_contents(operation_group)))

    @staticmethod
    def get_result(content):
        if content.get('metadata'):
            return content['metadata']['operation_result']
        elif content.get('result'):
            return content['result']
        else:
            assert False, content

    @classmethod
    def from_operation_group(cls, operation_group: dict, **predicates):
        """ Initialize with operation group contents.

        :param operation_group: operation_group: {"branch": "B...", "contents": [...], ...} \
        :param predicates: filter contents using predicates `field=value`
        :rtype: List[OperationResult]
        """
        if not cls.is_applied(operation_group):
            raise RpcError.from_errors(cls.errors(operation_group))

        def dispatch(content):
            if content['kind'] == 'transaction':
                return cls.from_transaction(content)
            elif content['kind'] == 'origination':
                return cls.from_origination(content)
            else:
                return content

        contents = cls.get_contents(operation_group, **predicates)
        return list(map(dispatch, contents))

    @classmethod
    def from_origination(cls, content: dict):
        """ Initialize with origination content.

        :param content:
        :rtype: OperationResult
        """
        operation_result = cls.get_result(content)
        return cls(
            storage=content['script']['storage'],
            originated_contracts=operation_result['originated_contracts']
        )

    @classmethod
    def from_transaction(cls, content: dict):
        """ Initialize with transaction content.

        :param content:
        :rtype: OperationResult
        """
        operation_result = cls.get_result(content)
        return cls(
            parameters=content.get('parameters'),
            storage=operation_result.get('storage'),
            big_map_diff=operation_result.get('big_map_diff', []),
            # TODO: if it is already an internal operation, we should think... (build a tree?)
            operations=cls.get_contents(content, source=content['destination'])
        )



import pendulum
from pendulum.parsing.exceptions import ParserError
from datetime import datetime
from itertools import count
from typing import Iterator

from pytezos.tools.docstring import get_attr_docstring
from pytezos.rpc.search import BlockSliceQuery
from pytezos.rpc.query import RpcQuery
from pytezos.encoding import is_bh, is_ogh


def to_timestamp(v):
    try:
        v = pendulum.parse(v)
    except ParserError:
        pass
    if isinstance(v, datetime):
        v = int(v.timestamp())
    return v


class BlocksQuery(RpcQuery, path='/chains/{}/blocks'):

    def __call__(self, length=1, head=None, min_date=None):
        """ Lists known heads of the blockchain sorted with decreasing fitness.
        Optional arguments allows to returns the list of predecessors for known heads
        or the list of predecessors for a given list of blocks.

        :param length: The requested number of predecessors to returns (per requested head).
        :param head: An empty argument requests blocks from the current heads. \
        A non empty list allow to request specific fragment of the chain.
        :param min_date: When `min_date` is provided, heads with a timestamp before `min_date` are filtered out
        :rtype: list[list[str]]
        """
        if isinstance(head, str) and not is_bh(head):
            head = self.__getitem__(head).calculate_hash()

        if min_date and not isinstance(min_date, int):
            min_date = to_timestamp(min_date)

        return super(BlocksQuery, self).__call__(
            length=length, head=head, min_date=min_date)

    def __getitem__(self, block_id):
        """ Construct block query or get a block range.

        :param block_id: Block identity or block range

          - int: Block level or offset from the head if negative;
          - str: Block hash (base58) or special names (head, genesis), expressions like `head~1` etc;
          - slice [:]: First value (start) must be int, second (stop) can be any Block ID or empty.
        :rtype: BlockQuery or BlockSliceQuery
        """
        if isinstance(block_id, slice):
            if not isinstance(block_id.start, int):
                raise NotImplementedError('Slice start should be an integer.')

            return BlockSliceQuery(
                start=block_id.start,
                stop=block_id.stop,
                node=self.node,
                path=self._wild_path,
                params=self._params
            )

        if isinstance(block_id, int) and block_id < 0:
            return self.blocks[f'head~{block_id}']

        return super(BlocksQuery, self).__getitem__(block_id)

    @property
    def current_voting_period(self):
        """ Get block range for the current voting period.

        :rtype: BlockSliceQuery
        """
        metadata = self.head.metadata()
        return BlockSliceQuery(
            start=metadata['level']['level'] - metadata['level']['voting_period_position'],
            stop='head',
            node=self.node,
            path=self._wild_path,
            params=self._params
        )

    @property
    def current_cycle(self):
        """ Get block range for the current cycle.

        :rtype: BlockSliceQuery
        """
        metadata = self.head.metadata()
        return BlockSliceQuery(
            start=metadata['level']['level'] - metadata['level']['cycle_position'],
            stop='head',
            node=self.node,
            path=self._wild_path,
            params=self._params
        )


class BlockQuery(RpcQuery, path='/chains/{}/blocks/{}'):

    def __init__(self, *args, **kwargs):
        super(BlockQuery, self).__init__(*args, **kwargs)
        self._caching = self._caching or 'head' not in self._params

    @property
    def predecessor(self):
        """ Query previous block.

        :rtype: BlockQuery
        """
        return self._parent[self.header()['predecessor']]

    @property
    def baker(self):
        """ Query block producer (baker).

        :rtype: ContractQuery
        """
        return self.context.contracts[self.metadata()['baker']]

    def voting_period(self) -> int:
        """ Get voting period for this block from metadata.
        """
        return self.metadata()['level']['voting_period']

    def level(self) -> int:
        """ Get level for this block from metadata.
        """
        return self.metadata()['level']['level']

    def cycle(self) -> int:
        """
        Get cycle for this block from metadata.
        """
        return self.metadata()['level']['cycle']


class ContractQuery(RpcQuery, path='/chains/{}/blocks/{}/context/contracts/{}'):

    def public_key(self) -> str:
        """ Retrieve the contract manager's public key (base58 encoded)
        """
        pkh = self._params[-1]
        if pkh.startswith('KT'):
            pkh = self.manager_key().get('manager')

        pk = self._parent[pkh].manager_key().get('key')
        if not pk:
            raise ValueError('Public key is not revealed.')

        return pk

    def count(self) -> Iterator:
        """ Get contract counter iterator: it returns incremented value on each call.
        """
        return count(start=int(self.counter()) + 1, step=1)

    def code(self):
        """ Get contract code.

        :returns: Micheline expression
        """
        return self().get('script', {}).get('code')


class BigMapGetQuery(RpcQuery, path='/chains/{}/blocks/{}/context/contracts/{}/big_map_get'):

    def post(self, query: dict):
        """ Access the value associated with a key in the big map storage of the michelson.

        :param query
            {
                key: { $key_type : <key> },
                type: { "prim" : $key_prim }
            }
            $key_type: Provided key encoding, e.g. "string", "bytes" for hex-encoded string, "int"
            key_prim: Expected high-level data type, e.g. "address", "nat", "mutez" (see storage section in code)
        :returns: Micheline expression
        """
        return self._post(json=query)


class ContextRawBytesQuery(RpcQuery, path='/chains/{}/blocks/{}/context/raw/bytes'):

    def __init__(self, *args, **kwargs):
        kwargs.update(timeout=60)
        super(ContextRawBytesQuery, self).__init__(*args, **kwargs)

    def __call__(self, depth=1) -> dict:
        """ Return the raw context.

        :param depth: Context is a tree structure, default depth is 1
        """
        return super(ContextRawBytesQuery, self).__call__(depth=depth)


class ContextRawJsonQuery(RpcQuery, path='/chains/{}/blocks/{}/context/raw/json'):

    def __init__(self, *args, **kwargs):
        kwargs.update(timeout=60)
        super(ContextRawJsonQuery, self).__init__(*args, **kwargs)


class ContextSeedQuery(RpcQuery, path='/chains/{}/blocks/{}/context/seed'):

    def post(self):
        """ Get seed of the cycle to which the block belongs.
        """
        return self._post()


class EndorsingPower(RpcQuery, path='/chains/{}/blocks/{}/endorsing_power'):

    def post(self, endorsement_operation):
        """ Get the endorsing power of an endorsement operation, that is, the number of slots that the op has.

        :param endorsement_operation
            {
                "branch": $block_hash,
                "contents": [ $operation.alpha.contents ... ],
                "signature": $Signature
            }
        """
        return self._post({'sendorsement_operation': endorsement_operation})


class OperationListListQuery(RpcQuery, path=['/chains/{}/blocks/{}/operations']):

    def __getitem__(self, item):
        """ Find operation by hash.

        :param item: Operation group hash (base58)
        :rtype: OperationQuery
        """
        if isinstance(item, tuple):
            return self[item[0]][item[1]]

        if isinstance(item, str) and is_ogh(item):
            operation_hashes = self._parent.operation_hashes()

            def find_index():
                for i, validation_pass in enumerate(operation_hashes):
                    for j, og_hash in enumerate(validation_pass):
                        if og_hash == item:
                            return i, j
                raise StopIteration('Operation group hash not found')

            return self[find_index()]

        return super(OperationListListQuery, self).__getitem__(item)

    @property
    def endorsements(self):
        """ Operations with content of type: `endorsement`.

        :rtype: OperationListQuery
        """
        return self[0]

    @property
    def votes(self):
        """ Operations with content of type: `proposal`, `ballot`.

        :rtype: OperationListQuery
        """
        return self[1]

    @property
    def anonymous(self):
        """ Operations with content of type: `seed_nonce_revelation`, `double_endorsement_evidence`,
            `double_baking_evidence`, `activate_account`.

        :rtype: OperationListQuery
        """
        return self[2]

    @property
    def managers(self):
        """ Operations with content of type: `reveal`, `transaction`, `origination`, `delegation`.

        :rtype: OperationListQuery
        """
        return self[3]

    def find_upvotes(self, proposal_id) -> list:
        """ Find operations of kind `proposal` for given proposal.

        :param proposal_id: Proposal hash (base58)
        :returns: list of operation contents
        """
        def is_upvote(op):
            return any(map(
                lambda x: x['kind'] == 'proposal' and proposal_id in x.get('proposals', []),
                op['contents']))
        return list(filter(is_upvote, self.votes()))

    def find_ballots(self, proposal_id=None) -> list:
        """ Find operations of kind `ballot`.

        :param proposal_id: Proposal hash (optional)
        :returns: list of operation contents
        """
        def is_ballot(op):
            return any(map(
                lambda x: x['kind'] == 'ballot' and (proposal_id is None or proposal_id == x.get('proposal')),
                op['contents']))
        return list(filter(is_ballot, self.votes()))

    def find_origination(self, contract_id):
        """ Find origination of the contract.

        :param contract_id: Contract ID (KT-address)
        :returns: operation content
        """
        def is_origination(op):
            def is_it(x):
                return x['kind'] == 'origination' \
                       and contract_id in x['metadata']['operation_result']['originated_contracts']
            return any(map(is_it, op['contents']))
        return next(filter(is_origination, self.managers()))


class OperationQuery(RpcQuery, path=['/chains/{}/blocks/{}/operations/{}/{}']):

    def unsigned(self) -> dict:
        """ Get operation group data without metadata and signature.
        """
        data = self()
        return {
            'branch': data['branch'],
            'contents': [
                {k: v for k, v in content.items() if k != 'metadata'}
                for content in data['contents']
            ]
        }


class ProposalQuery(RpcQuery, path='/chains/{}/blocks/{}/votes/proposals/{}'):

    def __call__(self) -> int:
        """ Roll count for this proposal.
        """
        proposals = self._parent()
        proposal_id = self._params[-1]
        roll_count = next((x[1] for x in proposals if x[0] == proposal_id), 0)
        return roll_count


class ProposalsQuery(RpcQuery, path='/chains/{}/blocks/{}/votes/proposals'):

    def __getitem__(self, proposal_id) -> ProposalQuery:
        """ Roll count for the selected proposal.

        :param proposal_id: Base58-encoded proposal ID
        :rtype: ProposalQuery
        """
        return ProposalQuery(
            path=self._wild_path + '/{}',
            params=self._params + [proposal_id],
            node=self.node
        )

    def __repr__(self):
        res = [
            super(ProposalsQuery, self).__repr__(),
            f'[]{get_attr_docstring(self.__class__, "__getitem__")}',
        ]
        return '\n'.join(res)
